import time
from ddbpy.client import Send

dfe = ('127.0.0.1', 5555)

bucket = 'messages'

with Send(dfe) as send:
    send.switch_streaming(bucket)
    while True:
        ts = int(time.time())
        value = "123.456"
        metric = "float"
        send.send_payload(metric, ts, value)
        time.sleep(1)



from .messaging import Messaging

# for reading CSV
import csv

def toCsv(msg, nameColumn=True, timeColumn=False):
    def add_param(v):
        if v.strip() == '':
            v = '"%s"' % v
        if ',' in v and not((v.startswith('"') and v.endswith('"')) or (v.startswith("'") and v.endswith("'"))):
            v = '"%s"' % v
        return v + ", "
    ret = ""
    if timeColumn:
        t = msg.hdr.GetTime()
        if Messaging.findFieldInfo(msg.hdr.fields, "Time").units == "ms":
            t = t / 1000
        ret += str(t) + ', '
    if nameColumn:
        ret += msg.MsgName() + ", "
    for fieldInfo in Messaging.MsgClass(msg.hdr).fields:
        if(fieldInfo.count == 1):
            if not fieldInfo.exists(msg):
                break
            if len(fieldInfo.bitfieldInfo) == 0:
                columnText = add_param(str(Messaging.get(msg, fieldInfo)))
            else:
                columnText = ""
                for bitInfo in fieldInfo.bitfieldInfo:
                    columnText += add_param(str(Messaging.get(msg, bitInfo)))
        else:
            columnText = ""
            for i in range(0,fieldInfo.count):
                if not fieldInfo.exists(msg, i):
                    break
                columnText += add_param(str(Messaging.get(msg, fieldInfo, i)))
        ret += columnText
    # drop last two chars (', ') off end
    return ret[:-2]

def csvHeader(msg, nameColumn=True, timeColumn=False):
    tableHeader = ''
    if timeColumn:
        timeUnits = Messaging.findFieldInfo(msg.hdr.fields, "Time").units
        if timeUnits == "ms":
            timeUnits = "s"
        if timeUnits:
            timeUnits = " (%s)" % timeUnits
        tableHeader = "Time"+timeUnits+", "
    if nameColumn:
        tableHeader += msg.MsgName() + ", "
    for fieldInfo in type(msg).fields:
        if len(fieldInfo.bitfieldInfo) == 0:
            tableHeader += fieldInfo.name + ", "
        else:
            for bitInfo in fieldInfo.bitfieldInfo:
                tableHeader += bitInfo.name + ", "
    # drop last two chars (', ') off end
    return tableHeader[:-2]

def escapeCommasInQuotedString(line):
    ret = ""
    quoteStarted = 0
    for c in line:
        if c == '"':
            quoteStarted = not quoteStarted
        elif c == ',':
            if quoteStarted:
                ret = ret + '\\'
        ret = ret + c
    return ret

def msgNameAndParams(lineOfText):
    # check for message name followed by space
    params = lineOfText.split(" ", 1)
    msgName = params[0]
    if msgName in Messaging.MsgClassFromName:
        pass
    else:
        params = lineOfText.split(",", 1)
        msgName = params[0].strip()
    if len(params) == 1:
        params = []
    else:
        # escape commas that are inside quoted strings
        line = escapeCommasInQuotedString(params[1])
        # use CSV reader module
        params = list(csv.reader([line], quotechar='"', delimiter=',', quoting=csv.QUOTE_NONE, skipinitialspace=True, escapechar='\\'))[0]
        #print("params is " + str(params))
    return msgName, params
    
def csvToMsg(lineOfText):
    if lineOfText == '':
        return None
    msgName, params = msgNameAndParams(lineOfText)
    if msgName in Messaging.MsgClassFromName:
        msgClass = Messaging.MsgClassFromName[msgName]
        msg = msgClass()
        terminateMsg = 0
        terminationLen = 0
        if msg.fields:
            try:
                paramNumber = 0
                for fieldInfo in msgClass.fields:
                    val = params[paramNumber].strip()
                    #print("val is [" + val + "]") 
                    if(fieldInfo.count == 1):
                        if len(fieldInfo.bitfieldInfo) == 0:
                            if val.endswith(";"):
                                terminateMsg = 1
                                val = val[:-1]
                                if val == "":
                                    # terminate without this field
                                    terminationLen = fieldInfo.offset
                                    break
                                # terminate after this field
                                #TODO Broken for arrays-of-structs acting like parallel arrays!
                                terminationLen = fieldInfo.end_location()
                            if fieldInfo.type == "string":
                                if val.startswith('"') and val.endswith('"'):
                                    val = val.strip('"')
                                if terminateMsg:
                                    #TODO Broken for arrays-of-structs acting like parallel arrays!
                                    terminationLen = fieldInfo.end_location(len(val)-1)
                            Messaging.set(msg, fieldInfo, val)
                            paramNumber+=1
                        else:
                            for bitInfo in fieldInfo.bitfieldInfo:
                                if val.endswith(";"):
                                    terminateMsg = 1
                                    val = val[:-1]
                                    # terminate without anything after our parent field
                                    terminationLen = fieldInfo.end_location()
                                    if val == "":
                                        break
                                Messaging.set(msg, bitInfo, val)
                                paramNumber+=1
                                val = params[paramNumber]
                                if terminateMsg:
                                    break
                    else:
                        if val.startswith("0x") and len(val) > 2+2*fieldInfo.size:
                            if val.endswith(";"):
                                terminateMsg = 1
                                val = val[:-1]
                                hexStr = val[2:].strip()
                                #TODO Broken for arrays-of-structs acting like parallel arrays!
                                terminationLen = int(fieldInfo.offset + len(hexStr)/2)
                            hexStr = val[2:].strip()
                            charsForOneElem = fieldInfo.size*2
                            valArray = [hexStr[i:i+charsForOneElem] for i in range(0, len(hexStr), charsForOneElem)]
                            for i in range(0,len(valArray)):
                                Messaging.set(msg, fieldInfo, int(valArray[i], 16), i)
                            paramNumber+=1
                        else:
                            for i in range(0,fieldInfo.count):
                                if val.endswith(";"):
                                    terminateMsg = 1
                                    #TODO Broken for arrays-of-structs acting like parallel arrays!
                                    terminationLen = fieldInfo.end_location(i)
                                    val = val[:-1]
                                Messaging.set(msg, fieldInfo, val, i)
                                if terminateMsg:
                                    break
                                paramNumber+=1
                                val = params[paramNumber]
                    if terminateMsg:
                        break
            except IndexError:
                # if index error occurs on accessing params, then stop processing params
                # because we've processed them all
                pass
        if terminateMsg:
            msg.hdr.SetDataLength(terminationLen)
        return msg
    else:
        #print("["+lineOfText+"] is NOT A MESSAGE NAME!")
        pass
    return None

def long_substr(string_array):
    if len(string_array) == 0:
        return ''
    if len(string_array) == 1:
        return string_array[0]
    if len(string_array[0]) == 0:
        return string_array[0]
    substr = ''
    # iterate over chars in first string
    for c in range(len(string_array[0])):
        # check if it matches all other strings
        for s in range(1,len(string_array)):
            if len(string_array[s]) <= c or string_array[0][c] != string_array[s][c]:
                return substr
        substr += string_array[0][c]
    return substr

def paramHelp(field):
    ret = field.name
    if field.units:
        ret = ret + "(" + field.units + ")"
    if field.description:
        ret = ret + "# " + field.description
    return ret

def csvHelp(lineOfText):
    autoComplete = None
    help = ""
    msgName, params = msgNameAndParams(lineOfText)
    # if there's no params beyond first word, try to auto-complete a message name
    if len(params) == 0 and not lineOfText.endswith(" ") and not lineOfText.endswith(","):
        # search for messages that match us
        matchingMsgNames = []
        truncated = False
        for aMsgName in sorted(Messaging.MsgClassFromName.keys()):
            if aMsgName.startswith(msgName):
                # truncate to first dot after match
                firstdot = aMsgName.find('.',len(msgName))
                if firstdot > 0:
                    aMsgName = aMsgName[0:firstdot+1]
                    truncated = True
                if not matchingMsgNames or aMsgName != matchingMsgNames[-1]:
                    matchingMsgNames.append(aMsgName)
        if len(matchingMsgNames) == 1:
            help = matchingMsgNames[0]
            autoComplete = matchingMsgNames[0]
            # if there's only one result and we don't match it exactly (because it's longer than us)
            # accept it by giving it as autoComplete with a comma at end
            #if autoComplete != msgName:
            if not truncated:
                autoComplete = autoComplete + ','
            return (autoComplete, help)
        else:
            help = '\n'.join(matchingMsgNames)
            autoComplete = long_substr(matchingMsgNames)
            #print("long_substr returned " + autoComplete)
            return (autoComplete, help)
            
    if msgName in Messaging.MsgClassFromName:
        helpOnJustParam = len(params)>1 or (len(params) == 1 and params[0] != '')
        if not helpOnJustParam:
            help = msgName + ", "
        msgClass = Messaging.MsgClassFromName[msgName]
        msg = msgClass()
        if msg.fields:
            try:
                paramNumber = 0
                for fieldInfo in msgClass.fields:
                    if(fieldInfo.count == 1):
                        if len(fieldInfo.bitfieldInfo) == 0:
                            if helpOnJustParam:
                                if paramNumber == len(params):
                                    return (None, paramHelp(fieldInfo))
                                paramNumber+=1
                            else:
                                help += fieldInfo.name + ", "
                        else:
                            for bitInfo in fieldInfo.bitfieldInfo:
                                if helpOnJustParam:
                                    if paramNumber == len(params):
                                        return (None, paramHelp(bitInfo))
                                    paramNumber+=1
                                else:
                                    help += bitInfo.name + ", "
                    else:
                        if helpOnJustParam:
                            arrayList = []
                            for i in range(0,fieldInfo.count):
                                if helpOnJustParam and paramNumber == len(params):
                                    return (None, paramHelp(fieldInfo))
                                paramNumber+=1
                        else:
                            help += fieldInfo.name + "["+str(fieldInfo.count)+"], "
            except IndexError:
                # if index error occurs on accessing params, then stop processing params
                # because we've processed them all
                print("done at index " + str(paramNumber))
                pass
            if help.endswith(", "):
                help = help[:-2]
    else:
        return (None, "["+msgName+"] is not a message name!")
    return (autoComplete, help)



from datetime import datetime
import inspect
import sys
import traceback
import signal

def trace_lines(frame, event, arg):
    """Handler that executes with every line of code"""

    # We only care about *line* and *return* events
    if event != 'line' and event != 'return':
        return

    # Get a reference to the code object and source
    co = frame.f_code
    #source = inspect.getsourcelines(co)[0]

    # Send the UI information on the code we're currently executing
    trace_lines.applicationq.put(
        {"co": {"file": co.co_filename,
                "name": co.co_name,
                "lineno": frame.f_lineno},
         #"frame": {"lineno": frame.f_lineno,
         #          "firstlineno": co.co_firstlineno,
         #          "locals": str(frame.f_locals),
         #          "source": source},
         'trace': 'line'})

    # Wait for a debug command
    cmd = trace_lines.debugq.get()

    if cmd == "step":
        # If stepping through code, return this handler
        return trace_lines

    elif cmd == 'over':
        # If stepping over, then return nothing, unless we're at top level
        if co.co_name == "<module>":
            return trace_lines
        # If stepping out of code, return the function callback
        return trace_calls

    # print("CODE")
    # print("co_argcount " + str(co.co_argcount))
    # print("co_cellvars " + str(co.co_cellvars))
    # print("co_code " + str(co.co_code))
    # print("co_consts " + str(co.co_consts))
    # print("co_filename " + str(co.co_filename))
    # print("co_firstlineno " + str(co.co_firstlineno))
    # print("co_flags " + str(co.co_flags))
    # print("co_freevars " + str(co.co_freevars))
    # print("co_kwonlyargcount " + str(co.co_kwonlyargcount))
    # print("co_lnotab " + str(co.co_lnotab))
    # print("co_name " + str(co.co_name))
    # print("co_names " + str(co.co_names))
    # print("co_nlocals " + str(co.co_nlocals))
    # print("co_stacksize " + str(co.co_stacksize))
    # print("co_varnames " + str(co.co_varnames))
    #
    # print("FRAME")
    # print("clear " + str(frame.clear))
    # # print("f_back " + str(frame.f_back))
    # # print("f_builtins " + str(frame.f_builtins))
    # # print("f_code " + str(frame.f_code))
    # # print("f_globals " + str(frame.f_globals))
    # print("f_lasti " + str(frame.f_lasti))
    # print("f_lineno " + str(frame.f_lineno))
    # print("f_locals " + str(frame.f_locals))
    # print("f_trace " + str(frame.f_trace))

def trace_calls(frame, event, arg):
    """Handler that executes on every invocation of a function call"""

    # We only care about function call events
    if event != 'call':
        return

    # Get a reference for the code object and function name
    co = frame.f_code

    #TODO: If filename doesn't match, maybe still handle stepping,
    # as long as its source we might want to debug?
    if co.co_filename == trace_lines.debug_filename:
        # Get the source code from the code object
        #print("filename: " + co.co_filename)
        #source = inspect.getsourcelines(co)[0]

        # Tell the UI to perform an update
        trace_lines.applicationq.put(
            { "co": {"file": co.co_filename,
                      "name": co.co_name,
                      "lineno": frame.f_lineno},
              #"frame": {"lineno": frame.f_lineno,
              #          "firstlineno": co.co_firstlineno,
              #          "locals": str(frame.f_locals),
              #          "source": source},
              "trace": "call"})

        # Wait for a debug command (we stop here right before stepping into or out of a function)
        cmd = trace_lines.debugq.get()

        if cmd == 'step':
            # If stepping into the function, return the line callback
            return trace_lines
        elif cmd == 'over':
            # If stepping over, then return nothing, unless we're at top level
            if co.co_name == "<module>":
                return trace_lines
            return

    return

class stdout_capture:
    def write(self, data):
        trace_lines.applicationq.put({'stdout': data})
class stderr_capture:
    def write(self, data):
        trace_lines.applicationq.put({'stderr': data})

sout = stdout_capture()
serr = stderr_capture()

def exit_gracefully(signum, frame):
    exit()

def debug(applicationq, debugq, filename):
    """Sets up and starts the debugger"""

    # Setup the debug and application queues as properties of the trace_lines functions
    trace_lines.debugq = debugq
    trace_lines.applicationq = applicationq
    trace_lines.debug_filename = filename

    # capture output, to send to the applicationq
    sys.stdout = sout
    sys.stderr = serr

    # Enable debugging by setting the callback
    sys.settrace(trace_calls)
    
    # try to exit gracefully
    signal.signal(signal.SIGINT, exit_gracefully)
    signal.signal(signal.SIGTERM, exit_gracefully)

    # Execute the function we want to debug with its parameters
    name = filename.replace("/", "_")
    import importlib
    try:
        module = importlib.machinery.SourceFileLoader(name, filename).load_module(name)
        print("Finished")
        trace_lines.applicationq.put({'exit': True})
    except SystemExit:
        print("Stopped")
        trace_lines.applicationq.put({'error': True})
    except:
        trace_lines.applicationq.put({'error': True})
        etype, value, tb = sys.exc_info()
        exc = ''.join(traceback.format_exception(etype, value, tb))
        trace_lines.applicationq.put({'exception': exc})
        #trace_lines.applicationq.put({'exception': {'type': str(etype), 'value': str(value), 'traceback': str(tb)}})



#!/usr/bin/env python3
import os
import sys
import argparse
import traceback
import pkg_resources

# if started via invoking this file directly (like would happen with source sitting on disk),
# insert our relative msgtools root dir into the sys.path, so *our* msgtools is used, not
# any other already in the path.
if __name__ == '__main__':
    srcroot=os.path.abspath(os.path.dirname(os.path.abspath(__file__))+"/../..")
    sys.path.insert(1, srcroot)
from msgtools.lib.messaging import Messaging

from PyQt5 import QtCore, QtGui, QtWidgets, QtNetwork

from msgtools.server.TcpServer import *
from msgtools.server.WebSocketServer import *

DESCRIPTION='''
    MsgServer acts as a central routing hub for one or more message clients.
    Messages received, are reflected onto all other client connections.
    Bluetooth (RFCOMM), TCP, Websocket, and USB Serial clients are all
    supported.  

    MsgServer also provides binary logging capability from from both the UI
    and Network.xLogx messages.  Message masks and subscription are also
    supported through subscription messages defined in the Network 
    group.
'''

EPILOG='''

Headers
=======
MsgServer expects headers.NetworkHeader as the prequel for all messages.  
All clients need to send and receive messages with a NetworkHeader.

TCP and Websockets
==================
MsgServer always runs a TCP and Websocket server.  The Websocket server
port is always one greater than the TCP port.  The default ports are
5678 and 5679 for TCP and Websockets respectively.

SPP and RFCOMM
==============
SPP ("Serial Port Profile") is a Bluetooth service profile for emulating 
an RS232 serial connection over Bluetooth. To accomplish this, SPP uses an 
RFCOMM (a reliable stream-oriented connection; the TCP of the Bluetooth 
world) socket to move the data.  SPP provides RS-232 signaling concepts
for CTS, RTS, etc.

BluetoothRFCOMMQt
=================
* Requires Qt
* Works on Mac and Linux.

BluetoothRFCOMM
===============
* Requires PyBlueZ which may not be available for Python versions after 3.5
* Works on Windows and Linux

Bluetooth Arguments
===================
Both command line options take a Bluetooth MAC address, and optionally 
an RFCOMM port to connect to. For instance: 

    --bluetoothRFCOMM=00:de:ad:be:ef:77,8

If you omit the port (",8", above), then the code will attempt to use 
Bluetooth SDP to find an SPP service on the target device; if multiple 
services are available, an arbitrary one will be chosen.

Bluetooth Troubleshooting
=========================
If you can't connect to your device, try pairing it with your PC first.
Sometimes devices require pairing before allowing an arbitrary connection.

Examples
========
    msgserver &

    Will run MsgServer on the default TCP Websocket ports 5678 and 5679.

    msgserver --serial

    Will run MsgServer with a serial port input on the last used port, 
    and TCP/Websocket on 5678, and 5679 respectively.

    msgserver --serial= /dev/tty12

    Will run MsgServer with a serial port connected on /dev/tty12. and
    a TCP/Websocket port on 5678, and 5679 respectively.

    msgserver --port 1234

    Will run MsgServer with a TCP port on 1234 and Websocket port on 1235

    msgserver --bluetoothRFCOMM=00:de:ad:be:ef:77

    Will run MsgServer on the default TCP/Websocket ports and attempt to connect
    to the indicated Bluetooth MAC.  SDP will be used to try and discover
    the appropriate channel.


    msgserver --bluetoothRFCOMM=00:de:ad:be:ef:77,8

    Will run MsgServer on the default TCP/Websocket ports and attempt to connect
    to the indicated Bluetooth MAC and channel 8.

'''

class SelectPluginDialog(QtWidgets.QDialog):
    pluginSelected = QtCore.pyqtSignal(str)
    def __init__(self, parent=None):
        super(SelectPluginDialog, self).__init__(parent)
        self.setWindowModality(QtCore.Qt.ApplicationModal)
        self.setWindowTitle("Select a PLugin")
        
        self.resize(600, 200)

        self.pluginList = QtWidgets.QTreeWidget()
        openButton = QtWidgets.QPushButton("Load")
        openButton.clicked.connect(self.loadPlugin)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.pluginList)
        layout.addWidget(openButton)
        self.setLayout(layout)

        tableHeader = ["Name", "Module"]
        self.pluginList.setHeaderLabels(tableHeader)
        for entry_point in pkg_resources.iter_entry_points("msgtools.server.plugin"):
            try:
                plugin_entry_point = entry_point.load()
                try:
                    pluginInfo = plugin_entry_point
                    if pluginInfo.enabled():
                        list = [pluginInfo.name, entry_point.module_name]
                        item = QtWidgets.QTreeWidgetItem(None, list)
                        item.plug_entrypoint_name = entry_point.name
                        self.pluginList.addTopLevelItem(item)
                    else:
                        if Messaging.debug:
                            print("Ignoring plugin %s %s:%s is disabled" % (entry_point.name, entry_point.module_name, entry_point.attrs[0]))
                except AttributeError as e:
                    if Messaging.debug:
                        print("    " + str(e))
                        print("No info in %s %s:%s!  must be function" % (entry_point.name, entry_point.module_name, entry_point.attrs[0]))
                    list = [entry_point.name, entry_point.module_name]
                    item = QtWidgets.QTreeWidgetItem(None, list)
                    item.plug_entrypoint_name = entry_point.name
                    self.pluginList.addTopLevelItem(item)
            except Exception as e:
                if Messaging.debug:
                    print(e)
                    import traceback
                    print(traceback.format_exc())                    
                    print("Ignoring plugin %s %s:%s, exception while loading" % (entry_point.name, entry_point.module_name, entry_point.attrs[0]))
        for i in range(0, len(tableHeader)):
            self.pluginList.resizeColumnToContents(i)
        
    def loadPlugin(self):
        cur_item = self.pluginList.currentItem()
        if cur_item is not None:
            self.pluginSelected.emit(cur_item.plug_entrypoint_name)
            self.close()

class MessageServer(QtWidgets.QMainWindow):
    def __init__(self):
        QtWidgets.QMainWindow.__init__(self)
        
        self.settings = QtCore.QSettings("MsgTools", "MessageServer")
        self.logFile = None
        self.logFileType = None
        
        parser = argparse.ArgumentParser(description=DESCRIPTION, epilog=EPILOG, 
            formatter_class=argparse.RawDescriptionHelpFormatter)

        # iterate over plugins, and setup argparse for each
        for entry_point in pkg_resources.iter_entry_points("msgtools.server.plugin"):
            parser.add_argument(
                '--%s' %entry_point.name,
                dest='last%s'%entry_point.name,
                action='store_true',
                help='''Load the %s plugin via %s:%s''' % (entry_point.name, entry_point.module_name, entry_point.attrs[0]))
            parser.add_argument(
                '--%s='%entry_point.name,
                dest= entry_point.name,
                help="Same as above, passing param %s to the plugin" % (entry_point.name.upper()))

        parser.add_argument('--plugin', help='Specify a filesystem path to a plugin module to load.')
        parser.add_argument('--msgdir', 
            help='''Specify the directory for generated python code to use for headers, and other 
                    message definitions''')
        parser.add_argument('--port', type=int, 
            help='The TCP port to use.  Websockets are always TCP port + 1.')
        parser.add_argument('--debug', action='store_true', help='Set if you want extra error info printed to stdout.')

        # if we had plugins before, but no cmdline args now, add simulated
        # cmdline args for our old plugins
        if len(sys.argv) == 1:
            last_plugins = self.settings.value("pluginsLoaded", "", str).split("|")
            for plugin in last_plugins:
                if plugin:
                    sys.argv.append("--"+plugin)

        args = parser.parse_args()
        
        Messaging.debug = False if hasattr(args, 'debug') == False else args.debug
        
        try:
            Messaging.LoadAllMessages(searchdir=args.msgdir)
        except ImportError:
            print("\nERROR! Auto-generated python code not found!")
            print("cd to a directory downstream from a parent of obj/CodeGenerator/Python")
            print("or specify that directory with --msgdir=PATH\n")
            quit()

        self.networkMsgs = Messaging.Messages.Network

        self.clients = {}
        
        self.privateSubscriptions = {}

        self.initializeGui()

        self.pluginPorts = []
        tcpport = 5678
        wsport = 5679

        if args.port is not None:
            tcpport = args.port
            wsport = tcpport+1
        
        for entry_point in pkg_resources.iter_entry_points("msgtools.server.plugin"):
            # check for argparse data for the plugin
            plugin_option = getattr(args, entry_point.name)
            plugin_last_option = getattr(args, 'last'+entry_point.name)
            if plugin_last_option is not False or plugin_option is not None:
                param = plugin_option if plugin_option is not None else None
                self.load_plugin(entry_point.name, entry_point, param)

        # load plugin via path to .py file
        if args.plugin is not None:
            filename = args.plugin
            import os
            moduleName = os.path.splitext(os.path.basename(filename))[0]
            if Messaging.debug:
                print("loading module %s as %s" % (filename, moduleName))

            name = filename.replace("/", "_")
            import importlib
            self.plugin = importlib.machinery.SourceFileLoader(name, filename).load_module(name)

            pluginPort = self.plugin.PluginConnection(None)
            pluginPort.statusUpdate.connect(self.onStatusUpdate)
            # try to allow plugin to emit new connections
            try:
                pluginPort.newConnection.connect(self.onNewConnection)
            except AttributeError:
                # if that fails, just use the plugin as one new connection.
                self.onNewConnection(pluginPort)
            pluginPort.start()
            self.pluginPorts.append(pluginPort)

        self.tcpServer = TcpServer(tcpport)
        self.tcpServer.statusUpdate.connect(self.onStatusUpdate)
        self.tcpServer.newConnection.connect(self.onNewConnection)

        self.wsServer = WebSocketServer(wsport)
        self.wsServer.statusUpdate.connect(self.onStatusUpdate)
        self.wsServer.newConnection.connect(self.onNewConnection)

        self.tcpServer.start()
        self.wsServer.start()
        name = self.tcpServer.serverInfo() + "(TCP) and " + str(self.wsServer.portNumber) + "(WebSocket)"
        self.statusBar().addPermanentWidget(QtWidgets.QLabel(name))
        self.readSettings()

    def initializeGui(self):
        # Layout
        vbox = QtWidgets.QVBoxLayout()

        # Components

        # if started via invoking this file directly (like would happen with source sitting on disk),
        # do not show the button to load plugins, since that only works with installed plugins.
        if __name__ != "__main__":
            self.addPluginButton = QtWidgets.QPushButton("Load Plugin")
            self.addPluginButton.pressed.connect(self.onAddPluginClicked)
            vbox.addWidget(self.addPluginButton)

        self.logButton = QtWidgets.QPushButton("Start Logging")
        self.logButton.pressed.connect(self.onLogButtonClicked)
        vbox.addWidget(self.logButton)

        self.grid = QtWidgets.QGridLayout()
        vbox.addLayout(self.grid)
        
        self.statusBox = QtWidgets.QPlainTextEdit()
        vbox.addWidget(self.statusBox)

        # Central Widget (QMainWindow limitation)
        centralWidget = QtWidgets.QWidget()
        centralWidget.setLayout(vbox)
        self.setCentralWidget(centralWidget)

        # Main Window Stuff
        self.setWindowTitle("MessageServer 0.1")
        self.statusBar()

    def startLog(self, log_name):
        if self.logFile:
            self.logFile.close()
            self.logFile = None
        if log_name.endswith("json"):
            self.logFileType = "json"
        elif log_name.endswith("csv"):
            self.logFileType = "csv"
            # hash table of booleans for if each type of message has had it's header
            # put into this log file already.
            self.loggedMsgHeaderRow = {}
        elif log_name.endswith('bin') or log_name.endswith('log'):
            self.logFileType = "bin"
        else:
            print("ERROR!  Invalid log type " + log_name)
        self.logFile = QtCore.QFile(log_name)
        self.logFile.open(QtCore.QIODevice.Append)
        fileInfo = QtCore.QFileInfo(log_name)
        self.settings.setValue("logging/filename", fileInfo.dir().absolutePath())
        self.logButton.setText("Stop " + fileInfo.fileName())
        self.queryLog()
    
    def stopLog(self):
        if self.logFile != None:
            self.logFile.close()
            self.logFile = None
            self.logButton.setText("Start Logging")
            self.queryLog()

    def queryLog(self):
        if hasattr(self.networkMsgs, 'LogStatus'):
            logStatusMsg = self.networkMsgs.LogStatus()
            if self.logFile != None:
                logStatusMsg.SetLogOpen(1)
                logStatusMsg.SetLogFileName(self.logFile.fileName())
                if self.logFileType == "json":
                    logStatusMsg.SetLogFileType("JSON")
                elif self.logFileType == "csv":
                    logStatusMsg.SetLogFileType("CSV")
                elif self.logFileType == "bin":
                    logStatusMsg.SetLogFileType("BIN")
            for client in self.clients.values():
                client.sendMsg(logStatusMsg.hdr)

    def onLogButtonClicked(self):
        if self.logFile != None:
            self.stopLog()
        else:
            currentDateTime = QtCore.QDateTime.currentDateTime()
            defaultFilename = currentDateTime.toString("yyyyMMdd-hhmmss") + ".log"
            logFileName, _ = QtWidgets.QFileDialog.getSaveFileName(self, "Save File", self.settings.value("logging/filename", ".")+"/"+defaultFilename, "Log Files (*.log);;JSON Files (*.json);;CSV Files (*.csv)")
            # if they hit cancel, don't do anything
            if not logFileName:
                return
            self.startLog(logFileName)
    
    def onAddPluginClicked(self):
        d = SelectPluginDialog()
        d.pluginSelected.connect(self.pluginSelected)
        d.exec_()
    
    def pluginSelected(self, plugin_name):
        for entry_point in pkg_resources.iter_entry_points("msgtools.server.plugin"):
            if entry_point.name == plugin_name:
                self.load_plugin(plugin_name, entry_point, None)
                return

    def load_plugin(self, plugin_name, entry_point, param):
        try:
            # load the module, create the plugin
            plugin_entry_point = entry_point.load()
            try:
                # check for the entry point giving a struct with a 'connection' attribute
                pluginCreatorFn = plugin_entry_point.connect_function
            except AttributeError:
                pluginCreatorFn = plugin_entry_point
            pluginPort = pluginCreatorFn()
            pluginPort.plugin_name = plugin_name
        except:
            a,b,c = sys.exc_info()
            exc = ''.join(traceback.format_exception(a,b,c))
            self.onStatusUpdate("Exception loading plugin:\n%s" % (exc))
            print(exc)
            return
        
        # connect plugin to our status update function
        pluginPort.statusUpdate.connect(self.onStatusUpdate)
        
        # try to allow plugin to emit new connections
        try:
            pluginPort.newConnection.connect(self.onNewConnection)
        except AttributeError:
            # if that fails, just use the plugin as one new connection.
            self.onNewConnection(pluginPort)
        
        pluginPort.start()
        self.pluginPorts.append(pluginPort)

    def onStatusUpdate(self, message):
        self.statusBox.appendPlainText(message)

    def onNewConnection(self, newConnection):
        self.onStatusUpdate("adding connection[" + newConnection.name+"]")
        self.clients[newConnection] = newConnection
        newConnection.messagereceived.connect(self.onMessageReceived)
        newConnection.disconnected.connect(self.onConnectionDied)
        clientRow = self.grid.rowCount()
        i = 0
        while(1):
            widget = newConnection.widget(i)
            if widget == None:
                break
            self.grid.addWidget(widget, clientRow, i)
            i+=1
    
    def onConnectionDied(self, connection):
        self.onStatusUpdate("removing connection[" + connection.name+"]")
        # remove the plugin, if it is one
        try:
            self.pluginPorts.remove(connection)
        except ValueError:
            pass
        i = 0
        while(1):
            widget = connection.widget(i)
            if widget == None:
                break
            self.grid.removeWidget(widget)
            widget.deleteLater()
            i+=1
        if connection in self.clients:
            connection.deleteLater()
            del self.clients[connection]
        else:
            self.onStatusUpdate("cnx not in list!")

    def logMsg(self, hdr):
        #write to log, if log is open
        if self.logFile != None:
            log = ''
            if self.logFileType == "csv":
                msg = Messaging.MsgFactory(hdr)
                if not msg.MsgName() in self.loggedMsgHeaderRow:
                    self.loggedMsgHeaderRow[msg.MsgName()] = True
                    log = msg.csvHeader(timeColumn=True)+'\n'
                log += msg.toCsv(timeColumn=True)+'\n'
                log = log.encode('utf-8')
            elif self.logFileType == "json":
                msg = Messaging.MsgFactory(hdr)
                log = msg.toJson(includeHeader=True)+'\n'
                log = log.encode('utf-8')
            elif self.logFileType == "bin":
                log = hdr.rawBuffer().raw
            self.logFile.write(log)
            self.logFile.flush()

    def onMessageReceived(self, hdr):
        c = self.sender()
        # check for name, subscription, etc.
        if hasattr(self.networkMsgs, 'Connect') and hdr.GetMessageID() == self.networkMsgs.Connect.ID:
            connectMsg = self.networkMsgs.Connect(hdr.rawBuffer())
            c.name = connectMsg.GetName()
            c.statusLabel.setText(c.name)
        elif hasattr(self.networkMsgs, 'SubscriptionList') and hdr.GetMessageID() == self.networkMsgs.SubscriptionList.ID:
            c.subscriptions = {}
            subListMsg = self.networkMsgs.SubscriptionList(hdr.rawBuffer())
            for idx in range(0,self.networkMsgs.SubscriptionList.GetIDs.count):
                id = subListMsg.GetIDs(idx)
                if id != 0:
                    c.subscriptions[id] = id
            self.onStatusUpdate("updating subscription for "+c.name+" to " + ', '.join(hex(x) for x in c.subscriptions.keys()))
        elif hasattr(self.networkMsgs, 'MaskedSubscription') and hdr.GetMessageID() == self.networkMsgs.MaskedSubscription.ID:
            subMsg = self.networkMsgs.MaskedSubscription(hdr.rawBuffer())
            c.subMask = subMsg.GetMask()
            c.subValue = subMsg.GetValue()
            self.onStatusUpdate("updating subscription for "+c.name+" to id & " + hex(c.subMask) + " == " + hex(c.subValue))
        elif hasattr(self.networkMsgs, 'StartLog') and hdr.GetMessageID() == self.networkMsgs.StartLog.ID:
            startLog = self.networkMsgs.StartLog(hdr.rawBuffer())
            self.logFileType = startLog.GetLogFileType()
            logFileName = startLog.GetLogFileName()
            if not logFileName:
                logFileName = QtCore.QDateTime.currentDateTime().toString("yyyyMMdd-hhmmss") + ".log"
            self.startLog(logFileName)
        elif hasattr(self.networkMsgs, 'StopLog') and hdr.GetMessageID() == self.networkMsgs.StopLog.ID:
            self.stopLog()
        elif hasattr(self.networkMsgs, 'QueryLog') and hdr.GetMessageID() == self.networkMsgs.QueryLog.ID:
            self.queryLog()
        elif hasattr(self.networkMsgs, 'ClearLogs') and hdr.GetMessageID() == self.networkMsgs.ClearLogs.ID:
            # This message was added more for AndroidServer. Plenty of other good ways to delete a log on 
            # a desktop or other more capable machine. That said, silently eat the request - blindly forwarding
            # this along could be hazardous.
            pass
        elif hasattr(self.networkMsgs, 'Note') and hdr.GetMessageID() == self.networkMsgs.Note.ID:
            # The Note message allows a user to annotate a log.  We want to drop it into the log, but not
            # forward to all other clients.
            self.logMsg(hdr)

        elif hasattr(self.networkMsgs, 'PrivateSubscriptionList') and  hdr.GetMessageID() == self.networkMsgs.PrivateSubscriptionList.ID:
            subListMsg = self.networkMsgs.PrivateSubscriptionList(hdr.rawBuffer())
            privateSubs = []
            for idx in range(0,self.networkMsgs.PrivateSubscriptionList.GetIDs.count):
                id = subListMsg.GetIDs(idx)
                if id == 0:
                    break
                privateSubs.append(id)
                if id in self.privateSubscriptions:
                    self.privateSubscriptions[id].append(c)
                else:
                    self.privateSubscriptions[id] = [c]
            self.onStatusUpdate("adding Private subscription for "+c.name+": " + ', '.join(hex(x) for x in privateSubs))
        else:
            # Log the message
            self.logMsg(hdr)

            # Route to all clients
            for client in self.clients.values():
                if client != c:
                    id = hdr.GetMessageID()
                    if id in client.subscriptions or (id & client.subMask == client.subValue):
                        try:
                            if id in self.privateSubscriptions:
                                # if it's a "private" message, only give it to clients that specifically said they want it
                                # or to clients that are a hardware link.
                                if client in self.privateSubscriptions[id] or client.isHardwareLink:
                                    client.sendMsg(hdr)
                            else:
                                client.sendMsg(hdr)
                        except Exception as ex:
                            a,b,c = sys.exc_info()
                            exc = ''.join(traceback.format_exception(a,b,c))
                            self.onStatusUpdate("Exception in server.py while sending to client %s:\n%s" % (client.name, exc))
    def closeEvent(self, event):
        self.settings.setValue("geometry", self.saveGeometry())
        self.settings.setValue("windowState", self.saveState())
        pluginNames = []
        for pluginPort in self.pluginPorts:
            try:
                pluginNames.append(pluginPort.plugin_name)
            except AttributeError:
                pass
            pluginPort.stop()
        self.settings.setValue("pluginsLoaded", "|".join(pluginNames))
        super(MessageServer, self).closeEvent(event)

    def readSettings(self):
        self.restoreGeometry(self.settings.value("geometry", QtCore.QByteArray()))
        self.restoreState(self.settings.value("windowState", QtCore.QByteArray()))

def main(args=None):
    app = QtWidgets.QApplication(sys.argv)

    msgServer = MessageServer()
    msgServer.show()

    sys.exit(app.exec_())
    
if __name__ == "__main__":
    main()



#!/usr/bin/python3
import os

# if started via invoking this file directly (like would happen with source sitting on disk),
# insert our relative msgtools root dir into the sys.path, so *our* msgtools is used, not
# any other already in the path.
if __name__ == '__main__':
    srcroot=os.path.abspath(os.path.dirname(os.path.abspath(__file__))+"/../..")
    sys.path.insert(1, srcroot)
from msgtools.lib.messaging import Messaging
import msgtools.console.client

from influxdb import InfluxDBClient
from influxdb.client import InfluxDBClientError
import datetime

# this allows sending Messages as data to an InfluxDB database
class InfluxDBConnection:
    MAX_POINTS = 128
    def __init__(self, msgClient, dbname='messages', hostname='localhost', port=8086, username='root', password='root'):
        self.hostname = hostname
        self.port = port
        self.db = InfluxDBClient(host=hostname, port=port, username=username, password=password, database=dbname)
        self.msgClient = msgClient
        self.cookie = 0
        
        # History Data message has to have either just Data with interleaved value and time,
        # or be an array of structs which also interleaves value and time in memory, but
        # gives two functions to access them that each look like an array.
        self.dataInfo = Messaging.findFieldInfo(Messaging.Messages.Network.History.Data.fields, "Data")
        self.dataValueInfo = Messaging.findFieldInfo(Messaging.Messages.Network.History.Data.fields, "Data_Value")
        self.dataTimestampInfo = Messaging.findFieldInfo(Messaging.Messages.Network.History.Data.fields, "Data_Timestamp")
        if self.dataInfo != None:
            # multiply by 2 because the result includes a value and a timestamp
            self.dataResultSize = self.dataInfo.size *2
            self.dataResultOffset = self.dataInfo.offset
        else:
            # Size of a result includes size of the value and the timestamp.
            self.dataResultSize = self.dataValueInfo.size + self.dataTimestampInfo.size
            # This needs to be whichever element is first!
            self.dataResultOffset = min(self.dataValueInfo.offset+self.dataTimestampInfo.offset)
    
    @staticmethod
    def FormattedTime(floatTime):
        return str(int(floatTime * 1e9))

    @staticmethod
    def GetDBValue(msg, fieldInfo, index=0):
        # what to do for arrays?
        val = Messaging.get(msg, fieldInfo, index)
        if fieldInfo.type == 'int':
            val = int(val)
        elif fieldInfo.type == 'float':
            val = float(val)
        return val

    def handle_message(self, msg):
        if msg.MsgName().startswith("Network"):
            if msg.MsgName() == "Network.History.GetData":
                self.handle_query(msg)
        else:
            self.store_message(msg)
    
    def store_message(self, msg):
        try:
            timeVal = msg.hdr.GetTime()
            timeInfo = Messaging.findFieldInfo(msg.hdr.fields, "Time")
            maxTime = timeInfo.maxVal
            # if it's not big enough to be an absolute timestamp, give up on using it and just use current time
            if maxTime != "DBL_MAX" and (maxTime == 'FLT_MAX' or float(maxTime) <= 2**32):
                raise AttributeError
            if timeInfo.units == "ms":
                timeVal = timeVal / 1000.0
            timeVal = datetime.datetime.fromtimestamp(timeVal, datetime.timezone.utc)
        except AttributeError:
            timeVal = datetime.datetime.now()

        dbJson = {
                "time": str(timeVal),
                "measurement": msg.MsgName(),
                'fields':  {},
                'tags': {}
            }

        for fieldInfo in msg.hdr.fields:
            if len(fieldInfo.bitfieldInfo) == 0:
                if fieldInfo.idbits == 0 and fieldInfo.name != "Time" and fieldInfo.name != "DataLength":
                    dbJson['tags'][fieldInfo.name] = Messaging.get(msg.hdr, fieldInfo)
            else:
                for bitInfo in fieldInfo.bitfieldInfo:
                    if bitInfo.idbits == 0 and bitInfo.name != "Time" and bitInfo.name != "DataLength":
                        dbJson['tags'][bitInfo.name] = Messaging.get(msg.hdr, bitInfo)
        
        msgClass = type(msg)
        for fieldInfo in msgClass.fields:
            if fieldInfo.count == 1:
                if len(fieldInfo.bitfieldInfo) == 0:
                    dbJson['fields'][fieldInfo.name] = InfluxDBConnection.GetDBValue(msg, fieldInfo)
                else:
                    for bitInfo in fieldInfo.bitfieldInfo:
                        dbJson['fields'][bitInfo.name] = InfluxDBConnection.GetDBValue(msg, bitInfo)
            else:
                # flatten arrays
                for i in range(0,fieldInfo.count):
                    dbJson['fields'][fieldInfo.name+"_"+str(i)] = InfluxDBConnection.GetDBValue(msg, fieldInfo, i)

        #print("Create a retention policy")
        #retention_policy = 'awesome_policy'
        #client.create_retention_policy(retention_policy, '3d', 3, default=True)
        self.db.write_points([dbJson]) #, retention_policy=retention_policy)

    def handle_query(self, msg):
        msgquery = msg.GetQuery()
        msgAndFieldName = msgquery.split(",")[0]
        msgname = msgAndFieldName.split(".")[0]
        fieldname = msgAndFieldName.split(".")[1]
        tagQuery = msgquery.split(",")[1]
        if tagQuery != "":
            tagQuery += " AND "
        
        start = msg.GetStartTime()
        end = msg.GetEndTime()
        averagingPeriod = msg.GetAveragingPeriod()

        if averagingPeriod != 0:
            dbquery = 'SELECT "' + fieldname + '" FROM "' + msgname + '"'
        else:
            dbquery = 'SELECT MEAN("' + fieldname + '") FROM "' + msgname + '"'
        dbquery += " WHERE " + tagQuery + "time > " + FormattedTime(start)
        if end != 0:
            dbquery += " AND time < " + FormattedTime(end)
        else:
            end = datetime.datetime.now()
        if averagingPeriod != 0:
            # limit averaging period such that we don't get too many data points
            if avg / (end - start) > MAX_POINTS:
                avg = MAX_POINTS * (end - start)
            dbquery += " GROUP BY *,time("+FormattedTime(avg)+")"

        dbquery += " LIMIT " + str(MAX_POINTS)

        # send query to database, get results
        db_results = self.db.query(dbquery)

        # output results
        resultMsg = Messaging.Messages.Network.History.Result()
        resultMsg.SetCookie(self.cookie)
        resultCount = f(db_results)
        resultMsg.SetResultCount(resultCount)
        resultMsg.SetQuery(msgquery)
        self.msgClient.send_message(resultMsg)
        if self.dataInfo != None:
            maxDataPerMsg = self.dataInfo.count
        else:
            arrayOffset = self.dataTimestampInfo.offset
            arrayElemSize = self.dataValueInfo.size + dataTimestampInfo.size
            maxDataPerMsg = dataValueInfo.count
        resultMsgNumber = 0
        resultsLeftToSend = len(db_results)
        for result in db_results:
            dataMsg = Messaging.Messages.Network.History.Data()
            dataMsg.SetCookie(self.cookie)
            dataMsg.SetResultNumber(resultMsgNumber)
            dataRange = min(maxDataPerMsg, resultsLeftToSend)
            if self.dataInfo != None:
                for index in range(0, dataRange,2):
                    dataMsg.SetData(time(result), index)
                    dataMsg.SetData(value(result), index+1)
            else:
                for index in range(0, dataRange):
                    dataMsg.SetData_Timestamp(time(result), index)
                    dataMsg.SetData_Value(value(result), index)
            resultsLeftToSend -= dataRange
            resultMsgNumber += 1
            # truncate message buffer to actual amount of data
            msg_buffer_len = self.dataResultOffset + self.dataResultSize * dataRange
            self.msgClient.send_message(dataMsg)

        # Increment cookie so the next query's results are easy to tell apart from out query results
        # This is especially useful if two clients query at almost the same time (but only if they
        # have unique cookies!).
        self.cookie += 1

# this is client that reads from network, and writes to InfluxDB
class InfluxDBMsgClient:
    def __init__(self):
        Messaging.LoadAllMessages()

        self.connection = msgtools.console.client(Messaging.hdr)
        # say my name
        connectMsg = Messaging.Messages.Network.Connect()
        connectMsg.SetName("InfluxDB")
        self.connection.send_message(connectMsg)
        
        # do default subscription to get *everything*
        subscribeMsg = Messaging.Messages.Network.MaskedSubscription()
        self.connection.send_message(subscribeMsg)

        self.db = InfluxDBConnection(self)
    
    def run(self):
        while True:
            # this blocks until message received, or timeout occurs
            timeout = 10.0 # value in seconds
            hdr = self.connection.get_message(timeout)
            if hdr:
                msg = Messaging.MsgFactory(hdr)
                self.db.handle_message(msg)

# this is a CLI app that reads from network and writes to InfluxDB
def main(args=None):
    dbClient = InfluxDBMsgClient()
    try:
        dbClient.run()
    except KeyboardInterrupt:
        print('You pressed Ctrl+C!')
        dbClient.connection.stop()

if __name__ == "__main__":
    main()



import pkg_resources
import collections

def info():
    LauncherInfo = collections.namedtuple('LauncherInfo', ['icon_text', 'program_name', 'icon_filename'])
    return LauncherInfo('noisemaker', 'msgnoisemaker', pkg_resources.resource_filename('msgtools', 'noisemaker/noisemaker.png'))



#!/usr/bin/env python3
import os
import sys
import struct
from datetime import datetime
import collections
import functools
import argparse

from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *

# if started via invoking this file directly (like would happen with source sitting on disk),
# insert our relative msgtools root dir into the sys.path, so *our* msgtools is used, not
# any other already in the path.
if __name__ == '__main__':
    srcroot=os.path.abspath(os.path.dirname(os.path.abspath(__file__))+"/../..")
    sys.path.insert(1, srcroot)
from msgtools.lib.messaging import Messaging

import msgtools.lib.gui
import msgtools.lib.msgcsv as msgcsv
import msgtools.debug.debug

import msgtools.lib.txtreewidget as txtreewidget
plottingLoaded=0
try:
    from msgtools.lib.msgplot import MsgPlot
    plottingLoaded=1
except ImportError as e:
    print("Error loading plot interface ["+str(e)+"]")
    print("Perhaps you forgot to install pyqtgraph.")
except RuntimeError as e:
    print("Error loading plot interface ["+str(e)+"]")
    print("Perhaps you need to install the PyQt5 version of pyqtgraph.")
    
DESCRIPTION='''MsgScope provides a graphical interface that allow syou to view, plot, and send
    messages.  It requires defined messages in Python.  A list of discovered messages will be
    displayed in the upper left of the UI.'''

def slim_vbox(a, b, c=None):
    vLayout = QVBoxLayout()
    vLayout.setSpacing(0)
    vLayout.setContentsMargins(0,0,0,0)
    vLayout.addWidget(a)
    if c == None:
        vLayout.addWidget(b)
    else:
        hLayout = QHBoxLayout()
        hLayout.setSpacing(0)
        hLayout.setContentsMargins(0,0,0,0)
        hLayout.addWidget(b)
        hLayout.addWidget(c)
        vLayout.addLayout(hLayout)

    vBox = QWidget()
    vBox.setLayout(vLayout)
    return vBox

def vsplitter(parent, *argv):
    splitter = QSplitter(parent)
    splitter.setOrientation(Qt.Vertical)
    for arg in argv:
        splitter.addWidget(arg)
    return splitter

class ClosableDockWidget(QDockWidget):
    def __init__(self, name, parent, widget, itemList, itemList2):
        super(QDockWidget,self).__init__(name, parent)
        self.setObjectName(name)
        self.setWidget(widget)
        # list it needs to be removed from
        self.itemList = itemList
        self.itemList2 = itemList2

    def closeEvent(self, ev):
        self.itemList.remove(self.widget())
        self.itemList2.remove(self.widget())
        self.parent().removeDockWidget(self)

class MessageScopeGui(msgtools.lib.gui.Gui):
    def __init__(self, args, parent=None):
        msgtools.lib.gui.Gui.__init__(self, "Message Scope 0.1", args, parent)

        # event-based way of getting messages
        self.RxMsg.connect(self.ProcessMessage)

        self.configure_gui(parent, args.debugdicts)
        
        self.ReadTxDictionary()

    def configure_gui(self, parent, debugdicts):
        # create widgets for tx
        self.txDictionary = self.configure_tx_dictionary(parent)
        self.txMsgs = self.configure_tx_messages(parent)
        txClearBtn = QPushButton("Clear")

        if debugdicts:
            debugdictList = debugdicts.split(",")
        else:
            debugdictList = []
        self.debugWidget = msgtools.debug.debug.MsgDebugWidget(debugdictList, parent)
        self.debugWidget.messageOutput.connect(self.SendMsg)
        self.debugWidget.autocompleted.connect(self.textAutocomplete)

        # tracking what reply to expect from a typed command
        self.expectedReply = None
        
        # add them to the tx layout
        txVBox = slim_vbox(self.txMsgs, txClearBtn)
        self.txSplitter = vsplitter(parent, self.txDictionary, txVBox, self.debugWidget)
        
        # create widgets for rx
        self.rx_message_list = self.configure_rx_message_list(parent)
        self.rx_messages_widget = self.configure_rx_messages_widget(parent)
        self.configure_msg_plots(parent)
        rxClearListBtn = QPushButton("Clear")
        rxSortListBtn = QPushButton("Sort")
        rxClearMsgsBtn = QPushButton("Clear")
        
        # add them to the rx layout
        rxMsgListBox = slim_vbox(self.rx_message_list, rxClearListBtn, rxSortListBtn)
        rxMsgsBox = slim_vbox(self.rx_messages_widget, rxClearMsgsBtn)
        self.rxSplitter = vsplitter(parent, rxMsgListBox, rxMsgsBox)

        # top level horizontal splitter to divide the screen
        self.hSplitter = QSplitter(parent)
        self.hSplitter.addWidget(self.txSplitter)
        self.hSplitter.addWidget(self.rxSplitter)

        self.setCentralWidget(self.hSplitter)
    
        # connect signals for 'clear' buttons
        txClearBtn.clicked.connect(self.clear_tx)
        rxClearListBtn.clicked.connect(self.clear_rx_list)
        rxSortListBtn.clicked.connect(self.sort_rx_list)
        rxClearMsgsBtn.clicked.connect(self.clear_rx_msgs)
        
    def configure_msg_plots(self, parent):
        # dict of plot by msg_key
        self.msgPlotsByKey = {}
        self.msgPlotList = []

    def configure_tx_dictionary(self, parent):
        txDictionary = QTreeWidget(parent)
        txDictionary.itemDoubleClicked.connect(self.onTxMessageSelected)
        txDictionary.setHeaderLabels(["Transmit Dictionary"])
        return txDictionary

    def configure_tx_messages(self, parent):
        txMsgs = QTreeWidget(parent)
        txMsgs.setColumnCount(4)
        
        txMsgsHeader = QTreeWidgetItem(None, ["Message", "Field", "Value", "Units", "Description"])
        
        txMsgs.setHeaderItem(txMsgsHeader)
        return txMsgs

    def configure_rx_message_list(self, parent):
        self.rx_msg_list = {}

        timer = QTimer(self)
        timer.timeout.connect(self.show_rx_msg_rates)
        timer.start(1000)

        rxMessageList = QTreeWidget(parent)
        rxMessageList.setColumnCount(3)
        rxMsgHeader = QTreeWidgetItem(None, [ "Name", "Last Received", "Rx Rate" ])
        rxMessageList.setHeaderItem(rxMsgHeader)
        rxMessageList.setContextMenuPolicy(Qt.CustomContextMenu)

        rxMessageList.itemDoubleClicked.connect(self.onRxListDoubleClicked)
        rxMessageList.customContextMenuRequested.connect(lambda position : self.onRxMessageContextMenuRequested(position, rxMessageList))
        return rxMessageList
    

    def configure_rx_messages_widget(self, parent):
        self.rx_msg_widgets = {}
        rxMessagesTreeWidget = QTreeWidget(parent)
        rxMessagesTreeWidget.setColumnCount(4)
        rxMessagesTreeWidget.setDragEnabled(1)
        rxMessagesTreeWidget.setContextMenuPolicy(Qt.CustomContextMenu)
        rxMsgsHeader = QTreeWidgetItem(None, ["Message", "Field", "Value", "Units", "Description"])
        rxMessagesTreeWidget.setHeaderItem(rxMsgsHeader)
        rxMessagesTreeWidget.itemDoubleClicked.connect(self.onRxMessageFieldSelected)
        rxMessagesTreeWidget.customContextMenuRequested.connect(lambda position : self.onRxMessageContextMenuRequested(position, rxMessagesTreeWidget))
        return rxMessagesTreeWidget

    def ReadTxDictionary(self):
        #print("Tx Dictionary:")
        for id in Messaging.MsgNameFromID:
            name = Messaging.MsgNameFromID[id]
            components = name.split('.')
            dirs = components[:-1]
            msgName = components[-1]
            
            parentWidget = self.txDictionary
            parentPath = ""
            for dir in dirs:
                # find the node that matches the directory we're looking for
                dirItemMatches = self.txDictionary.findItems(dir, Qt.MatchExactly | Qt.MatchRecursive, 0)
                foundMatch = False
                for dirItem in dirItemMatches:
                    try:
                        if parentPath == dirItem.parentPath:
                            parentWidget = dirItem
                            foundMatch = True
                            break
                    except AttributeError:
                        pass
                # if we didn't find the node for the directory, add it
                if not foundMatch:
                    newWidget = QTreeWidgetItem(parentWidget)
                    newWidget.setText(0, dir)
                    newWidget.parentPath = parentPath
                    parentWidget = newWidget
                parentPath += dir + "."
            msgItem = QTreeWidgetItem(parentWidget)
            msgItem.setText(0, msgName)
            msgItem.msgName = name
        self.txDictionary.sortByColumn(0, Qt.AscendingOrder)

    def onTxMessageSelected(self, txListWidgetItem):
        # directories have children but messages don't, so only add messages by verifying the childCount is zero
        if txListWidgetItem.childCount() == 0:
            messageName = txListWidgetItem.msgName
            # Always add to TX panel even if the same message class may already exist
            # since we may want to send the same message with different contents/header/rates.
            message_class = Messaging.MsgClassFromName[messageName]
            messageObj = message_class() # invoke constructor

            messageTreeWidgetItem = txtreewidget.EditableMessageItem(self.txMsgs, messageObj, None)
            messageTreeWidgetItem.qobjectProxy.send_message.connect(self.on_tx_message_send)

    def on_open(self):
        # need to handle fields of multiple messages on same plot
        plot_count = self.settings.beginReadArray("plots")
        for i in range(plot_count):
            self.settings.setArrayIndex(i)
            plot = None
            line_count = self.settings.beginReadArray("lines")
            for j in range(line_count):
                self.settings.setArrayIndex(j)
                msg_key = self.settings.value("key")
                fieldName = self.settings.value("fieldname")
                msg_id = msg_key.split(',')[-1]
                try:
                    msgName = Messaging.MsgNameFromID[msg_id]
                except KeyError:
                    print('Error!  msg_id ' + msg_id + ' is undefined!')
                    continue
                msgClass = Messaging.MsgClassFromName[msgName]
                try:
                    if plot == None:
                        plot = self.addPlot(msgClass, msg_key, fieldName)
                    else:
                        plot.addLine(msgClass, msg_key, fieldName)
                        self.registerPlotForKey(msg_key, plot)
                except MsgPlot.PlotError as e:
                    print(e)
            self.settings.endArray()
        self.settings.endArray()

        self.txSplitter.restoreState(self.settings.value("txSplitterSizes", self.txSplitter.saveState()));
        self.rxSplitter.restoreState(self.settings.value("rxSplitterSizes", self.rxSplitter.saveState()));
        self.hSplitter.restoreState(self.settings.value("hSplitterSizes", self.hSplitter.saveState()));
        self.debugWidget.textEntryWidget.restoreState(self.settings.value("cmdHistory", self.debugWidget.textEntryWidget.saveState()));

    def on_close(self):
        self.settings.remove("plots")
        self.settings.beginWriteArray("plots")
        plotList = ""
        i = 0
        for plot in self.msgPlotList:
            self.settings.setArrayIndex(i)
            self.settings.beginWriteArray("lines")
            j = 0
            for line in plot.lines:
                self.settings.setArrayIndex(j)
                self.settings.setValue('key', line.msgKey)
                self.settings.setValue('fieldname', "%s[%d]" % (line.fieldInfo.name, line.fieldSubindex))
                j += 1
            self.settings.endArray()
            i += 1
        self.settings.endArray()
        
        # save splitter sizes and command history
        self.settings.setValue("txSplitterSizes", self.txSplitter.saveState());
        self.settings.setValue("rxSplitterSizes", self.rxSplitter.saveState());
        self.settings.setValue("hSplitterSizes",  self.hSplitter.saveState());
        self.settings.setValue("cmdHistory", self.debugWidget.textEntryWidget.saveState());
            
    def on_tx_message_send(self, msg):
        if not self.connected:
            self.OpenConnection()
        text = msgcsv.toCsv(msg)
        self.debugWidget.textEntryWidget.addText(text + " -> Msg\n> ")
        self.debugWidget.textEntryWidget.addToHistory(text)
        self.SendMsg(msg)
    
    def registerPlotForKey(self, msg_key, plot):
        if not msg_key in self.msgPlotsByKey:
            self.msgPlotsByKey[msg_key] = []
        if not plot in self.msgPlotsByKey[msg_key]:
            self.msgPlotsByKey[msg_key].append(plot)

    def addPlot(self, msgClass, msg_key, fieldName):
        plotListForKey = []
        if msg_key in self.msgPlotsByKey:
            plotListForKey = self.msgPlotsByKey[msg_key]
        else:
            self.msgPlotsByKey[msg_key] = plotListForKey
        plotName = msgClass.MsgName()
        if plottingLoaded:
            msgPlot = MsgPlot(msgClass, msg_key, fieldName)
            self.msgPlotList.append(msgPlot)
            # add a dock widget for new plot
            dockWidget = ClosableDockWidget(plotName, self, msgPlot, plotListForKey, self.msgPlotList)
            self.addDockWidget(Qt.RightDockWidgetArea, dockWidget)
            # Change title when plot is paused/resumed
            msgPlot.Paused.connect(lambda paused: dockWidget.setWindowTitle(plotName+" (PAUSED)" if paused else plotName))
            msgPlot.AddLineError.connect(lambda s: QMessageBox.warning(self, "Message Scope", s))
            # callback to register a plot to listen to a message ID.
            # this is triggered by plotting new message fields on a plot (including by drag-and-drop)
            def register_plot_for_message_key(key):
                plot = self.sender()
                self.registerPlotForKey(key, plot)
            msgPlot.RegisterForMessage.connect(lambda key: register_plot_for_message_key(key))
            plotListForKey.append(msgPlot)
            return msgPlot
        
    def onRxMessageFieldSelected(self, rxWidgetItem):
        if isinstance(rxWidgetItem, txtreewidget.FieldItem) or isinstance(rxWidgetItem, txtreewidget.FieldArrayItem):
            fieldInfo = rxWidgetItem.fieldInfo
            msg_id = hex(rxWidgetItem.msg.hdr.GetMessageID())
            msg_key = ",".join(Messaging.MsgRoute(rxWidgetItem.msg)) + "," + msg_id
            try:
                msgPlot = self.addPlot(type(rxWidgetItem.msg), msg_key, rxWidgetItem.fieldName)
                if msgPlot:
                    msgPlot.addData(rxWidgetItem.msg)
            except MsgPlot.PlotError as e:
                QMessageBox.warning(self, "Message Scope", str(e))

    def onRxMessageContextMenuRequested(self, pos, parent):
        rxWidgetItem = parent.itemAt(pos)
        if not rxWidgetItem:
            return
        msg_name = rxWidgetItem.msg.MsgName()
        msg_id = hex(rxWidgetItem.msg.hdr.GetMessageID())
        msg_key = ",".join(Messaging.MsgRoute(rxWidgetItem.msg)) + "," + msg_id
        
        menu = QMenu("Context menu", self)
        inspectAction = QAction('&Inspect', self)
        inspectAction.msg_name = msg_name
        inspectAction.msg_key = msg_key
        menu.addAction(inspectAction)
        inspectAction.triggered.connect(self.openMsgInspector)
        # add needed actions
        menu.exec(rxWidgetItem.treeWidget().viewport().mapToGlobal(pos))
    
    def openMsgInspector(self):
        sender = self.sender()

        args = []
        if self.connectionName:
            args = args + ['--connectionName='+self.connectionName]
        # If we were started with a specific msgdir, pass that on to msginspector
        try:
            self.msgdir
            args.append("--msgdir")
            args.append(self.msgdir)
        except AttributeError:
            pass
        # add the name of the message to the argument list
        args.append('--msg')
        args.append(sender.msg_name)
        #TODO do something with key!  it filters by sender/route.
        #sender.msg_key
        
        # launch msginspector with all the right args
        proc = QProcess(self)
        proc.start('msginspector', args)

                
    def ProcessMessage(self, msg):
        self.debugWidget.ProcessMessage(msg)

        hdr = msg.hdr
        msg_id = hex(hdr.GetMessageID())

        msg_key = ",".join(Messaging.MsgRoute(msg)) + "," + msg_id
        
        self.display_message_in_rx_list(msg_key, msg)
        self.display_message_in_rx_tree(msg_key, msg)
        self.display_message_in_plots(msg_key, msg)

    def onRxListDoubleClicked(self, rxListItem):
        self.add_message_to_rx_tree(rxListItem.msg_key, rxListItem.msg)

    def display_message_in_rx_list(self, msg_key, msg):
        rx_time = datetime.now()

        if not msg_key in self.rx_msg_list:
            widget_name = msg.MsgName()
            msg_route = Messaging.MsgRoute(msg)
            if len(msg_route) > 0 and not(all ("0" == a for a in msg_route)):
                widget_name += " ("+",".join(msg_route)+")"
            msg_list_item = QTreeWidgetItem([ widget_name, rx_time.strftime('%H:%M:%S.%f')[:-3], "- Hz" ])
            msg_list_item.msg_key = msg_key
            msg_list_item.msg = msg
            msg_list_item.rx_count = 1
            msg_list_item.avg_rate = 1.0

            self.rx_message_list.addTopLevelItem(msg_list_item)
            self.rx_message_list.resizeColumnToContents(0)
            self.rx_msg_list[msg_key] = msg_list_item

        self.rx_msg_list[msg_key].setText(1, rx_time.strftime('%H:%M:%S.%f')[:-3])
        self.rx_msg_list[msg_key].msg = msg
        self.rx_msg_list[msg_key].rx_count += 1

    def show_rx_msg_rates(self):
        weight = 0.5
        for msg_key, widget in self.rx_msg_list.items():
            rate = float(widget.rx_count)
            widget.rx_count = 0
            widget.avg_rate = weight * widget.avg_rate + (1-weight) * rate

            if widget.avg_rate > 0.05:
                output = "{0:0.1f} Hz".format(widget.avg_rate)
            elif widget.avg_rate > 0.01:
                output = "0 Hz"
            else:
                output = "-- Hz"

            self.rx_msg_list[msg_key].setText(2, output)

    def add_message_to_rx_tree(self, msg_key, msg):
        if not msg_key in self.rx_msg_widgets:
            msg_widget = txtreewidget.MessageItem(self.rx_messages_widget, msg, msg_key)
            self.rx_msg_widgets[msg_key] = msg_widget
            self.rx_messages_widget.addTopLevelItem(msg_widget)
            self.rx_messages_widget.resizeColumnToContents(0)

    def display_message_in_rx_tree(self, msg_key, msg):
        if msg_key in self.rx_msg_widgets:
            self.rx_msg_widgets[msg_key].set_msg_buffer(msg.rawBuffer())
    
    def display_message_in_plots(self, msg_key, msg):
        if msg_key in self.msgPlotsByKey:
            plotListForKey = self.msgPlotsByKey[msg_key]
            for plot in plotListForKey:
                plot.addData(msg)
    
    def clear_rx_list(self):
        self.rx_msg_list = {}
        self.rx_message_list.clear()

    def sort_rx_list(self):
        self.rx_message_list.setSortingEnabled(not self.rx_message_list.isSortingEnabled())

    def clear_rx_msgs(self):
        self.rx_msg_widgets = {}
        self.rx_messages_widget.clear()

    def clear_tx(self):
        self.txMsgs.clear()
    
    # when text edit autocomplete occurs, find the corresponding item
    # in the tx dictionary
    def textAutocomplete(self, autocomplete):
        def item_matches(item, sequence):
            # look in reverse at sequence of parts, verifying
            # that each part matches the item and it's parents.
            for part in reversed(sequence):
                if part == '':
                    continue
                if item.text(0) != part:
                    return False
                item = item.parent()
            return True
        msgname_parts = autocomplete.replace(',', '').split('.')
        for msgpart in msgname_parts:
            # find all matches...
            matches = self.txDictionary.findItems(msgpart, Qt.MatchExactly | Qt.MatchRecursive, 0)
            for item in matches:
                # ... then for each match, make sure all parts match
                if item_matches(item, msgname_parts):
                    self.txDictionary.setCurrentItem(item)

def main():
    # Setup a command line processor...
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser = msgtools.lib.gui.Gui.addBaseArguments(parser)
    parser.add_argument('--debugdicts', help=''''Dictionaries to use for debug message format strings.''')
    args = parser.parse_args()

    app = QApplication(sys.argv)
    msgScopeGui = MessageScopeGui(args)
    msgScopeGui.show()
    sys.exit(app.exec_())

# main starts here
if __name__ == '__main__':
    main()



#!/usr/bin/env python3
#
# Client is a class that contains a TCP client which run asynchronously
# in a background thread, and also allows synchronous code to send and receives messages with it.
#
import os
import sys
import socket
import time
from msgtools.lib.messaging import Messaging

class Client:
    def __init__(self, name='Client', timeout=10):
        # keep a reference to Messages, for convenience of cmdline scripts
        self.Messages = Messaging.Messages
        # load all messages if not already loaded
        if Messaging.hdr == None:
            Messaging.LoadAllMessages()

        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._sock.connect(("127.0.0.1", 5678))
        self._timeout = timeout

        if self._timeout == 0.0:
            self._sock.setblocking(0)
        else:
            self._sock.setblocking(1)
            self._sock.settimeout(self._timeout)

        # say my name
        connectMsg = Messaging.Messages.Network.Connect()
        connectMsg.SetName(name)
        self.send(connectMsg)
        
        # do default subscription to get *everything*
        subscribeMsg = Messaging.Messages.Network.MaskedSubscription()
        self.send(subscribeMsg)
        
        # keep a dictionary of the latest value of all received messages
        self.received = {}
        
        # also record particular messages the user wants to record, even if
        # they never called recv on them
        self._extra_msgs_to_record = {}
    
    def send(self, msg):
        bufferSize = len(msg.rawBuffer().raw)
        computedSize = msg.hdr.SIZE + msg.hdr.GetDataLength()
        if(computedSize > bufferSize):
            msg.hdr.SetDataLength(bufferSize - msg.hdr.SIZE)
            print("Truncating message to "+str(computedSize)+" bytes")
        if(computedSize < bufferSize):
            # don't send the *whole* message, just a section of it up to the specified length
            self._sock.send(msg.rawBuffer().raw[0:computedSize])
        else:
            self._sock.send(msg.rawBuffer().raw)

    
    def recv(self, msgIds=[], timeout=None):
        # if user didn't pass a list, put the single param into a list
        if not isinstance(msgIds, list):
            msgIds = [msgIds]
        # if they passed classes, get the ID of each
        for i in range(0,len(msgIds)):
            if hasattr(msgIds[i], 'ID'):
                msgIds[i] = msgIds[i].ID
        if timeout != None and self._timeout != timeout:
            self._timeout = timeout
            if self._timeout == 0.0:
                self._sock.setblocking(0)
            else:
                self._sock.setblocking(1)
                self._sock.settimeout(self._timeout)
        while True:
            try:
                # see if there's enough for header
                data = self._sock.recv(Messaging.hdr.SIZE, socket.MSG_PEEK)
                if len(data) == Messaging.hdr.SIZE:
                    # create header based on peek'd data
                    hdr = Messaging.hdr(data)

                    # see if there's enough for the body, too
                    data += self._sock.recv(hdr.GetDataLength(), socket.MSG_PEEK)
                    if len(data) != Messaging.hdr.SIZE + hdr.GetDataLength():
                        print("didn't get whole body, error!")
                        continue

                    # read out what we peek'd.
                    data = self._sock.recv(Messaging.hdr.SIZE + hdr.GetDataLength())

                    # reset the header based on appended data
                    hdr = Messaging.hdr(data)
                    id = hdr.GetMessageID()
                    if id in self._extra_msgs_to_record:
                        self.received[id] = msg
                    if len(msgIds) == 0 or id in msgIds:
                        msg = Messaging.MsgFactory(hdr)
                        self.received[id] = msg
                        return msg
            except socket.timeout:
                return None
            except BlockingIOError:
                return None

    def wait_for(self, msgIds=[], fn=None, timeout=None):
        ret = False
        if timeout == None:
            timeout = self._timeout
        remaining_timeout = timeout
        while(1):
            before = time.time()
            msg = self.recv(msgIds, remaining_timeout)
            if msg != None and fn(msg):
                ret = True
                break
            after = time.time()
            elapsed = after - before
            if remaining_timeout:
                # if we have a timeout, update it for the amount of time spent
                remaining_timeout = remaining_timeout - elapsed
                if remaining_timeout < 0:
                    ret = False
                    break
        if timeout:
            self._sock.setblocking(1)
            self._sock.settimeout(self._timeout)
        return ret

    def stop(self):
        self._sock.close()

    # select set of messages to record the last value of
    def record(self, msgs_to_record):
        self._extra_msgs_to_record = msgs_to_record



#!/usr/bin/env python3
import sys, os
import argparse
from PyQt5 import QtCore, QtGui, QtWidgets
import pkg_resources
from . import codegen_gui

DESCRIPTION='''MsgLauncher launches msgtools applications.  It gives them relevant settings (like server
IP address and port) when it starts them.'''

class DetachableProcess(QtCore.QProcess):
    def __init__(self):
        QtCore.QProcess.__init__(self)
    def detach(self):
        self.waitForStarted()
        self.setProcessState(QtCore.QProcess.NotRunning);

    
class MsgLauncher(QtWidgets.QMainWindow):
    def __init__(self, parent=None):
        QtWidgets.QMainWindow.__init__(self,parent)
        
        parser = argparse.ArgumentParser(description=DESCRIPTION)
        parser.add_argument('--connectionName', default=None,
            help='''The connection name.  For socket connections this is an IP and port e.g. 127.0.0.1:5678.  
                    You may prepend ws:// to indicate you want to use a Websocket instead.  For file 
                    connection types, this is the filename to use as a message source.
                    This parameter is overridden by the --ip and --port options.''')
        parser.add_argument('--msgdir', help=''''The directory to load Python message source from.''')
        args = parser.parse_args()
        args.msgdir = None if hasattr(args, 'msgdir') == False else args.msgdir
        self.msgdir = args.msgdir

        # persistent settings
        self.settings = QtCore.QSettings("MsgTools", "launcher")
        
        self.restoreGeometry(self.settings.value("geometry", QtCore.QByteArray()))

        self.connectionName = self.settings.value("connection", "")
        
        if args.connectionName != None:
            self.connectionName = args.connectionName
        
        settingsAction = QtWidgets.QAction('&Settings', self)
        menubar = self.menuBar()
        connectMenu = menubar.addMenu('&Connection')
        connectMenu.addAction(settingsAction)
        settingsAction.triggered.connect(self.chooseHost)
        
        if os.path.exists('/opt/msgtools/messages'):
            self.codegen_gui = codegen_gui.CodegenGui(menubar, self.settings)
        
        self.setWindowTitle("MsgLauncher")
        
        # list of everything we launched
        self.procs = []
        
        # get list of programs to put into the launcher
        apps = self.programs_to_launch()
        
        # create a grid for launcher buttons
        grid = QtWidgets.QGridLayout()
        grid.setSpacing(0)
        grid.setContentsMargins(0,0,0,0)
        num_cols = 3
        col = 0
        row = 0
        for k in sorted(apps):
            app_info = apps[k]
            app_launcher = QtWidgets.QToolButton()
            app_launcher.setText(app_info.icon_text)

            # set up icon
            icon_filename = app_info.icon_filename
            if os.path.exists(icon_filename):
                pixmap = QtGui.QPixmap(icon_filename)
                icon = QtGui.QIcon(pixmap)
                app_launcher.setIcon(icon)
                app_launcher.setIconSize(pixmap.rect().size()/2)
                app_launcher.setFixedSize(pixmap.rect().size())
                app_launcher.setToolButtonStyle(QtCore.Qt.ToolButtonTextUnderIcon)
            else:
                print("ERROR!  %s does not exist!" % icon_filename)

            # set up launching when clicked
            app_launcher.program_name = app_info.program_name
            app_launcher.clicked.connect(self.launch)
            
            # add to layout
            grid.addWidget(app_launcher, row, col)
            
            # adjust column and row for next icon
            col += 1
            if col >= num_cols:
                col = 0
                row += 1
                
        w = QtWidgets.QWidget(self)
        w.setLayout(grid)
        self.setCentralWidget(w)
        self.adjustSize()
        self.setFixedSize(self.size())

    def programs_to_launch(self):
        # class to store info.  can't used collections.namedtuple because it's not mutable
        class LauncherInfo:
            def __init__(self, icon_text, program_name, icon_filename, module_name):
                self.icon_text = icon_text
                self.program_name = program_name
                self.icon_filename = icon_filename
                self.module_name = module_name

        # Make a list of programs to launch
        progs = []
        exclude_msgtools = False
        # check if we were launched from source on disk, or via an installed package entry point
        if __name__ == "__main__":
            # exclude anything from installed python packages
            exclude_msgtools = True

            # look for source near our file location on disk
            basedir = os.path.abspath(os.path.dirname(__file__)) + '/..'
            for filename in os.listdir(basedir):
                if filename == "launcher":
                    continue
                dirname = os.path.abspath(basedir + "/" + filename)
                png_name = dirname+'/'+filename+'.png'
                if os.path.exists(dirname+'/launcher.py') and os.path.exists(png_name):
                    progs.append(LauncherInfo(filename, [sys.executable, dirname+'/'+filename+'.py'], png_name, 'msgtools.'+filename+'.launcher'))

        # use pkg_resources to find programs to launch
        for entry_point in pkg_resources.iter_entry_points("msgtools.launcher.plugin"):
            launcher_info_fn = entry_point.load()
            launcher_info = launcher_info_fn()
            if exclude_msgtools and entry_point.module_name.startswith('msgtools.'):
                continue
            progs.append(LauncherInfo(*launcher_info, module_name=entry_point.module_name))

        # put the list into a hash table that has keys that can be sorted
        # alphabetically to give the order we want.
        ret = {}
        for launcher_info in progs:
            # come up with a sorted name to prioritize the core apps
            # above the non-core apps, even within the msgtools package.
            sort_name = launcher_info.icon_text
            if launcher_info.module_name.startswith('msgtools.'):
                sort_name = '@'+ sort_name
                if launcher_info.module_name.startswith('msgtools.server'):
                    sort_name = '@1'+ sort_name
                elif launcher_info.module_name.startswith('msgtools.scope'):
                    sort_name = '@2'+ sort_name
                elif launcher_info.module_name.startswith('msgtools.script'):
                    sort_name = '@3'+ sort_name
                elif (launcher_info.module_name.startswith('msgtools.debug') or
                      launcher_info.module_name.startswith('msgtools.inspector')):
                    sort_name = '@'+ sort_name
                else:
                    # any msgtools. items that *aren't* in the above if/elif, will
                    # sorted after those that are.
                    pass
            if len(launcher_info.program_name) == 2:
                launcher_info.icon_text = './' + launcher_info.icon_text
                sort_name = './' + sort_name
            ret[sort_name] = launcher_info

        return ret

    def launch(self):
        sender = self.sender()
        program_name = sender.program_name
        args = []
        is_msgserver = (program_name == 'msgserver')
        if len(program_name) == 2:
            args = [program_name[1]]
            program_name = program_name[0]
            is_msgserver = args[0].endswith('msgtools/server/server.py')
        if self.connectionName and not is_msgserver:
            args = args + ['--connectionName='+self.connectionName]
        if self.msgdir:
            args.append("--msgdir="+self.msgdir)

        proc = DetachableProcess()
        proc.finished.connect(self.process_exited)
        proc.readyReadStandardOutput.connect(self.printStdOut)
        proc.readyReadStandardError.connect(self.printStdErr)
        proc.start(program_name, args)
        self.procs.append(proc)

    def chooseHost(self):
        userInput, ok = QtWidgets.QInputDialog.getText(self, 'Connect',  'Server:', QtWidgets.QLineEdit.Normal, self.connectionName)
        if ok:
            self.connectionName = userInput
            self.settings.setValue("connection", self.connectionName)

    def process_exited(self, exitCode, exitStatus):
        p = self.sender()
        self.procs.remove(p)

    def closeEvent(self, event):
        self.settings.setValue("geometry", self.saveGeometry())
        self.settings.setValue("connection", self.connectionName)
        
        if len(self.procs) > 0:
            ret = QtWidgets.QMessageBox.warning(
                self,
                "MsgLauncher Exiting",
                "Close launched applications?",
                QtWidgets.QMessageBox.YesToAll | QtWidgets.QMessageBox.No);
            if ret == QtWidgets.QMessageBox.YesToAll:
                for p in self.procs:
                    p.terminate()
                    p.detach()
            elif ret == QtWidgets.QMessageBox.No:
                for p in self.procs:
                    p.detach()

        try:
            self.codegen_gui.save_settings()
        except AttributeError:
            pass

        super(QtWidgets.QMainWindow, self).closeEvent(event)
    
    def printStdOut(self):
        sender = self.sender()
        print(str(sender.readAllStandardOutput(), encoding='utf-8'))

    def printStdErr(self):
        sender = self.sender()
        print(str(sender.readAllStandardError(), encoding='utf-8'))

def main(args=None):
    app = QtWidgets.QApplication(sys.argv)
    msgApp = MsgLauncher()
    msgApp.show()
    sys.exit(app.exec_())

# main starts here
if __name__ == '__main__':
    main()



from PyQt5 import QtCore, QtGui, QtWidgets, QtNetwork
from PyQt5.QtCore import QObject

from msgtools.lib.messaging import Messaging
from BluetoothHeader import BluetoothHeader
from msgtools.lib.header_translator import HeaderTranslator

# We require Qt Bluetooth support, available on Linux and Macs(?)

class BluetoothRFCOMMQtConnection(QObject):
    disconnected = QtCore.pyqtSignal(object)
    messagereceived = QtCore.pyqtSignal(object)
    statusUpdate = QtCore.pyqtSignal(str)

    def __init__(self, socket):
        super(BluetoothRFCOMMQtConnection, self).__init__(None)

        self.removeClient = QtWidgets.QPushButton("Remove")
        self.removeClient.pressed.connect(self.onDisconnected)

        self.statusLabel = QtWidgets.QLabel()
        self.subscriptions = {}
        self.subMask = 0
        self.subValue = 0
        self.isHardwareLink = True

        self.socket = socket
        self.socket.readyRead.connect(self.onReadyRead)
        self.socket.disconnected.connect(self.onDisconnected)

        self.hdrTranslator = HeaderTranslator(BluetoothHeader, Messaging.hdr)
        
        self.rxBuffer = bytearray()

        self.name = "Bluetooth RFCOMM " + self.socket.peerAddress().toString()
        self.statusLabel.setText(self.name)

    def start(self):
        pass

    def widget(self, index):
        if index == 0:
            return self.removeClient
        if index == 1:
            return self.statusLabel
        return None
            
    def onReadyRead(self):
        inputStream = QtCore.QDataStream(self.socket)

        while(1):
            if len(self.rxBuffer) < BluetoothHeader.SIZE:
                if self.socket.bytesAvailable() < BluetoothHeader.SIZE:
                    return
                self.rxBuffer += inputStream.readRawData(BluetoothHeader.SIZE - len(self.rxBuffer))

            if len(self.rxBuffer) >= BluetoothHeader.SIZE:
                hdr = BluetoothHeader(self.rxBuffer)
                bodyLen = hdr.GetDataLength()
                if len(self.rxBuffer)+self.socket.bytesAvailable() < BluetoothHeader.SIZE + bodyLen:
                    return

                self.rxBuffer += inputStream.readRawData(BluetoothHeader.SIZE + bodyLen - len(self.rxBuffer))

                # create a new header object with the appended body
                btHdr = BluetoothHeader(self.rxBuffer)

                # if we got this far, we have a whole message! So, emit the signal
                networkMsg = self.hdrTranslator.translate(btHdr)

                self.messagereceived.emit(networkMsg)

                # then clear the buffer, so we start over on the next message
                self.rxBuffer = bytearray()

    def onDisconnected(self):
        self.socket.close()
        self.disconnected.emit(self)

    def sendMsg(self, networkMsg):
        btMsg = self.hdrTranslator.translate(networkMsg)
        # if we can't translate the message, just return
        if btMsg == None:
            return
        self.socket.write(btMsg.rawBuffer().raw)

    def stop(self):
        pass

def PluginConnection(param=""):
    from msgtools.server.BluetoothRFCOMMQt import BluetoothRFCOMMQtConnection
    from PyQt5 import QtBluetooth
    btArgs = param.split(",")
    btHost = btArgs[0]
    if len(btArgs)>1:
        btPort = int(btArgs[1])
    else:
        btPort = 8
    btSocket = QtBluetooth.QBluetoothSocket(QtBluetooth.QBluetoothServiceInfo.RfcommProtocol)
    btSocket.connectToService(QtBluetooth.QBluetoothAddress(btHost), btPort)
    return BluetoothRFCOMMQtConnection(btSocket)

def PluginEnabled():
    return True

import collections
PluginInfo = collections.namedtuple('PluginInfo', ['name', 'enabled', 'connect_function'])
plugin_info = PluginInfo('Bluetooth RF Comm (Qt)', PluginEnabled, PluginConnection)



#!/usr/bin/env python3
import os
import sys
import argparse
from PyQt5 import QtGui, QtWidgets, QtCore

# if started via invoking this file directly (like would happen with source sitting on disk),
# insert our relative msgtools root dir into the sys.path, so *our* msgtools is used, not
# any other already in the path.
if __name__ == '__main__':
    srcroot=os.path.abspath(os.path.dirname(os.path.abspath(__file__))+"/../..")
    sys.path.insert(1, srcroot)
from msgtools.lib.messaging import Messaging
import msgtools.lib.gui

DESCRIPTION='''
    Connects to a MsgSever and listens for BandwidthTest messages. When a message is received
    it is timestamped and sent back.  This app is meant to be used with the BandwidthTester app 
    which sources the BandwidthTest messages and displays results.  The UI also lets you adjust
    % of received messages that are simply dropped.
    '''

class BandwidthTestEcho(msgtools.lib.gui.Gui):
    def __init__(self, parent=None):
        parser = argparse.ArgumentParser(description=DESCRIPTION)
        parser = msgtools.lib.gui.Gui.addBaseArguments(parser)
        args = parser.parse_args()

        msgtools.lib.gui.Gui.__init__(self, "Bandwidth Test Echo 0.1", args, parent)
        
        self.msgCount = 0
        self.dropPercent = 0
        self.rxBytesPerSec = 0
        
        for msgClassName in Messaging.MsgClassFromName:
            if "BandwidthTest" in msgClassName:
                self.bandwidthTestMsgClass = Messaging.MsgClassFromName[msgClassName]
                break
        
        # event-based way of getting messages
        self.RxMsg.connect(self.ProcessMessage)

        vbox = QtWidgets.QVBoxLayout()
        centralWidget = QtWidgets.QWidget()
        centralWidget.setLayout(vbox)
        self.setCentralWidget(centralWidget)
        
        dropPercentSlider = QtWidgets.QSlider(QtCore.Qt.Horizontal, self)
        dropPercentSlider.valueChanged.connect(self.setDropPercent)
        dropPercentLabel = QtWidgets.QLabel()
        dropPercentSlider.valueChanged.connect(lambda newVal: dropPercentLabel.setText(str(newVal)+" % of msgs will be dropped"))
        dropPercentSlider.setMinimum(0)
        dropPercentSlider.setMaximum(100)
        hbox = QtWidgets.QHBoxLayout()
        vbox.addLayout(hbox)
        hbox.addWidget(QtWidgets.QLabel("Drop %"))
        hbox.addWidget(dropPercentSlider)
        hbox.addWidget(dropPercentLabel)

        hbox = QtWidgets.QHBoxLayout()
        vbox.addLayout(hbox)
        hbox.addWidget(QtWidgets.QLabel("Bytes/sec"))
        self.rxBytesPerSecLabel = QtWidgets.QLabel("")
        hbox.addWidget(self.rxBytesPerSecLabel)

        self.displayTimer = QtCore.QTimer(self)
        self.displayTimer.setInterval(1000)
        self.displayTimer.timeout.connect(self.updateDisplay)
        self.displayTimer.start()

    def setDropPercent(self, dropPercent):
        self.dropPercent = dropPercent

    def ProcessMessage(self, msg):
        if type(msg) == self.bandwidthTestMsgClass:
            self.msgCount += 1
            if self.msgCount >= self.dropPercent:
                self.sendBytesFn(msg.rawBuffer().raw)
                self.rxBytesPerSec += msg.hdr.GetDataLength()
            if self.msgCount >= 100:
                self.msgCount = 0

    def updateDisplay(self):
        self.rxBytesPerSecLabel.setText(str(self.rxBytesPerSec))
        self.rxBytesPerSec = 0

def main(args=None):
    app = QtWidgets.QApplication(sys.argv)
    msgApp = BandwidthTestEcho()
    msgApp.show()
    sys.exit(app.exec_())

# main starts here
if __name__ == '__main__':
    main()



from collections import namedtuple
import ctypes
import datetime
import struct

from .messaging import Messaging

class HeaderTranslator:
    def __init__(self, hdr1, hdr2):
        # Make a list of fields in the headers that have matching names.
        self._correspondingFields = []
        for fieldInfo1 in hdr1.fields:
            if len(fieldInfo1.bitfieldInfo) == 0:
                fieldInfo2 = Messaging.findFieldInfo(hdr2.fields, fieldInfo1.name)
                if fieldInfo2 != None:
                    self._correspondingFields.append([fieldInfo1, fieldInfo2])
            else:
                for bitfieldInfo1 in fieldInfo1.bitfieldInfo:
                    fieldInfo2 = Messaging.findFieldInfo(hdr2.fields, bitfieldInfo1.name)
                    if fieldInfo2 != None:
                        self._correspondingFields.append([bitfieldInfo1, fieldInfo2])
        
        HdrInfo = namedtuple('HdrInfo', 'type infoIndex timeField')
        self._hdr1Info = HdrInfo(hdr1, 0, Messaging.findFieldInfo(hdr1.fields, "Time"))
        self._hdr2Info = HdrInfo(hdr2, 1, Messaging.findFieldInfo(hdr2.fields, "Time"))
        self._timestampOffset = 0
        self._lastTimestamp = 0
        self.lastWrapTime = None

    def translateHdrAndBody(self, fromHdr, body):
        # figure out which direction to translate
        if isinstance(fromHdr, self._hdr1Info.type):
            fromHdrInfo = self._hdr1Info
            toHdrInfo = self._hdr2Info
        elif isinstance(fromHdr, self._hdr2Info.type):
            fromHdrInfo = self._hdr2Info
            toHdrInfo = self._hdr1Info
        else:
            print("ERROR!  type %s is not %s or %s!" % (type(fromHdr), self._hdr1Info.type, self._hdr2Info.type))
            raise TypeError
        
        # allocate the message to translate to
        toBuffer = ctypes.create_string_buffer(toHdrInfo.type.SIZE+fromHdr.GetDataLength())
        toHdr = toHdrInfo.type(toBuffer)
        toHdr.initialize()

        # loop through fields using reflection, and transfer contents from
        # one header to the other
        for pair in self._correspondingFields:
            fromFieldInfo = pair[fromHdrInfo.infoIndex]
            toFieldInfo = pair[toHdrInfo.infoIndex]
            Messaging.set(toHdr, toFieldInfo, Messaging.get(fromHdr, fromFieldInfo))
        
        # if the message ID can't be expressed in the new header, return None,
        # because this message isn't translatable
        if fromHdr.GetMessageID() != toHdr.GetMessageID():
            if Messaging.debug:
                print("message ID 0x" + hex(fromHdr.GetMessageID()) + " translated to 0x" + hex(toHdr.GetMessageID()) + ", throwing away")
            return None
        # copy the body
        for i in range(0,fromHdr.GetDataLength()):
            toHdr.rawBuffer()[toHdr.SIZE+i] = body[i]
        
        # do special timestamp stuff to convert from relative to absolute time
        if toHdrInfo.timeField != None:
            def set_time(hdr, t):
                # This is a bit ugly, but it's hard to tell with the Time field is a float or
                # an int.  If it's an int and we give it a float, struct.error gets raised.
                try:
                    hdr.SetTime(t)
                except struct.error:
                    hdr.SetTime(int(t))
            if fromHdrInfo.timeField != None:
                def timeReallyBig(tmax):
                    if tmax == 'DBL_MAX':
                        return True
                    if tmax == 'FLT_MAX':
                        return True
                    if int(tmax) >= 2**32:
                        return True
                def timeScaling(fromUnits, toUnits):
                    if fromUnits == 'ms' and toUnits == 's':
                        return 0.001
                    if fromUnits == 's' and toUnits == 'ms':
                        return 1000.0
                    return 1.0
                time_scale = timeScaling(fromHdrInfo.timeField.units, toHdrInfo.timeField.units)
                # if we're converting from a header with a smaller timestamp to a header
                # with a bigger timestamp, look for wrapping of the input timestamp
                time_could_wrap = False
                if not timeReallyBig(fromHdrInfo.timeField.maxVal):
                    if timeReallyBig(toHdrInfo.timeField.maxVal):
                        time_could_wrap = True
                    else:
                        time_could_wrap = int(fromHdrInfo.timeField.maxVal) < int(toHdrInfo.timeField.maxVal)
                if time_could_wrap:
                    # Detect time rolling
                    thisTimestamp = fromHdr.GetTime() * time_scale
                    thisTime = datetime.datetime.now()
                    if thisTimestamp < self._lastTimestamp:
                        # If the timestamp shouldn't have wrapped yet, assume messages sent out-of-order,
                        # and do not wrap again.
                        if (self.lastWrapTime == None or
                            thisTime > self.lastWrapTime + datetime.timedelta(0,30)):
                            self.lastWrapTime = thisTime
                            self._timestampOffset += 1
                    self._lastTimestamp = thisTimestamp
                    # need to handle different size timestamps!
                    set_time(toHdr, self._timestampOffset * (1+int(fromHdrInfo.timeField.maxVal)) + thisTimestamp)
                else:
                    set_time(toHdr, fromHdr.GetTime()*time_scale)
            else:
                t = datetime.datetime.now().timestamp()
                # use time since start of day, if 32-bit or smaller timestamps
                if float(toHdrInfo.timeField.maxVal) <= 2**32:
                    t = (datetime.datetime.fromtimestamp(t) - datetime.datetime.fromtimestamp(t).replace(hour=0, minute=0, second=0, microsecond=0)).total_seconds()
                if toHdrInfo.timeField.units == "ms":
                    t = t * 1000.0
                if toHdrInfo.timeField.type == "int":
                    t = int(t)
                set_time(toHdr, t)

        return toHdr

    def translate(self, fromHdr):
        toHdr = self.translateHdrAndBody(fromHdr, fromHdr.rawBuffer()[type(fromHdr).SIZE:])
        if toHdr != None:
            self.promoteIdFields(toHdr)
        return toHdr

    # Handle message body ID fields by populating them in message header.
    # This is to handle systems where some message body fields are conditionally
    # used as ID fields.
    def promoteIdFields(self, hdr):
        # if any message body fields are marked as IDs, look for header fields with
        # same name if header field is zero, set it to value of body field.
        id = hdr.GetMessageID()
        keep_going = True
        while(keep_going):
            keep_going = False
            msgClass = Messaging.MsgClass(hdr)
            msg = msgClass(hdr.rawBuffer())
            for fieldInfo in msg.fields:
                if fieldInfo.idbits != 0:
                    hdrFieldInfo = Messaging.findFieldInfo(hdr.fields, fieldInfo.name)
                    if hdrFieldInfo and hdrFieldInfo.get(hdr) == 0:
                        hdrFieldInfo.set(hdr, fieldInfo.get(msg))
                        keep_going = True



#!/usr/bin/env python3
import sys
import os
import math
import argparse
import signal

from PyQt5 import QtCore, QtGui, QtWidgets

# if started via invoking this file directly (like would happen with source sitting on disk),
# insert our relative msgtools root dir into the sys.path, so *our* msgtools is used, not
# any other already in the path.
if __name__ == '__main__':
    srcroot=os.path.abspath(os.path.dirname(os.path.abspath(__file__))+"/../..")
    sys.path.insert(1, srcroot)
from msgtools.lib.messaging import Messaging
import msgtools.lib.gui
import msgtools.lib.msgcsv as msgcsv

DESCRIPTION='''
    Lumberjack creates a subdirectory, and one CSV file per message type received in that directory.
    You may source data from any connectionType.  If you specify a logfile name positional
    Lumberjack assumes you want to source from a logfile.
'''
class Lumberjack(msgtools.lib.gui.Gui):
    def __init__(self, parent=None):

        parser = argparse.ArgumentParser(description=DESCRIPTION)
        parser.add_argument('-o', '--outputdir', help='''Specifies the name of the directory to output
            parsed data to.  Required for non-file connectionTypes.''')
        parser.add_argument('logfile', nargs='?', default=None, help='''The log file you want to split into CSV.  
            .log extension assumes the log was created by MsgServer (binary).  A .txt extension assumes the 
            file was created by SD logger.  This option is a pseudonym for --connectionType='file' and 
            --connectionName=<filename>, and will override connectionType, and connectionName.''')
        parser=msgtools.lib.gui.Gui.addBaseArguments(parser)
        args = parser.parse_args()

        # Special Handling here for files...
        if args.logfile is not None:
            args.connectionType = 'file'
            args.connectionName = args.logfile
            if args.logfile.lower().endswith('.txt'):
                args.serial = True
            args.ip = None
            args.port = None


        msgtools.lib.gui.Gui.__init__(self, "Lumberjack 0.1", args, parent)
        
        # If the user specified the output dir use it.  If the user 
        # specified a source file
        if args.outputdir is not None:
            self.outputName = args.outputdir
        elif args.connectionType == 'file':
            if args.connectionName is not None:
                self.outputName = self.connectionName.replace('.log', '')
                self.outputName = self.outputName.replace('.txt', '')
                self.outputName = self.outputName.replace('.TXT', '')
            else:
                print('''You must specify the name of the source file in --connectionName
                    of logfile for a \'file\' connectionType.''')
                sys.exit(1)
        else:
            print('You must specify the -o option if you aren\'t using a \'file\' connectionType')
            sys.exit(1)

        if os.path.exists(self.outputName) is False:
            os.makedirs(self.outputName)
        print("outputName is " + self.outputName + "\n")

        # event-based way of getting messages
        self.RxMsg.connect(self.ProcessMessage)
        
        self.outputFiles = {}

        # to handle timestamp wrapping
        self._timestampOffset = 0
        self._lastTimestamp = 0

        self.messageCount = 0

    def ProcessMessage(self, msg):
        self.messageCount += 1
        
        id = msg.hdr.GetMessageID()

        # if we write CSV to multiple files, we'd probably look up a hash table for this message id,
        # and open it and write a header
        if(id in self.outputFiles):
            outputFile = self.outputFiles[id]
        else:
            # create a new file
            outputFile = open(self.outputName + "/" + msg.MsgName().replace("/","_") + ".csv", 'w')

            # store a pointer to it, so we can find it next time (instead of creating it again)
            self.outputFiles[id] = outputFile
            
            # add table header, one column for each message field
            tableHeader = msgcsv.csvHeader(msg, nameColumn=False, timeColumn=True) + '\n'
            outputFile.write(tableHeader)
        
        try:
            # \todo Detect time rolling.  this only matters when we're processing a log file
            # with insufficient timestamp size, such that time rolls over from a large number
            # to a small one, during the log.
            thisTimestamp = msg.hdr.GetTime()
            if thisTimestamp < self._lastTimestamp:
                self._timestampOffset+=1

            self._lastTimestamp = thisTimestamp

            timeSizeInBits = int(round(math.log(int(Messaging.findFieldInfo(msg.hdr.fields, "Time").maxVal), 2)))
            timestamp = (self._timestampOffset << timeSizeInBits) + thisTimestamp
            if Messaging.findFieldInfo(msg.hdr.fields, "Time").units == "ms":
                timestamp = timestamp / 1000.0
            text = str(timestamp) + ", "
        except AttributeError:
            text = "unknown, "

        text += msgcsv.toCsv(msg, nameColumn=False, timeColumn=False)
        text += '\n'
        outputFile.write(text)

        # This is not efficient, but if we don't flush during socket processing
        # and the user hits Ctrl-C, we'll drop a bunch of data and end up with empty files.
        # So flush each message as it comes in.
        if self.connectionType !='file':
            outputFile.flush()

def main():
    app = QtWidgets.QApplication(sys.argv)
    msgApp = Lumberjack()

    # If we are processing a file then  use the Message Loop.  The reason for this dichotomy
    # in processing is the sockets are QtTcpSockets.  We've tied into Qt signals, which require
    # the app event loop to be running.  Files are handled with straight up Python files.

    # I can think of several other approaches, but they all pretty much boil down to sticking 
    # with QtApplication's event loop and going full on Qt, or kicking off a background worker 
    # thread and using Qt waitFor* constructs or straight up Python.

    # For now this little band-aid will get us by
    if msgApp.connectionType != 'file':
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        print('Listening for messages.  Press Ctrl-C to exit.')
        app.exec_()
    else:
        msgApp.MessageLoop()

    print("Processed " + str(msgApp.messageCount) + " messages")

# main starts here
if __name__ == '__main__':
    main()



import pkg_resources
import collections

def info():
    LauncherInfo = collections.namedtuple('LauncherInfo', ['icon_text', 'program_name', 'icon_filename'])
    return LauncherInfo('inspector', 'msginspector', pkg_resources.resource_filename('msgtools', 'inspector/inspector.png'))



#!/usr/bin/env python3
#
# Server is a class that contains a TCP server and Websocket server which run asynchronously
# in a background thread, and also allows synchronous code to send and receives messages with it.
#
# based on:
#   https://stackoverflow.com/questions/29324610/python-queue-linking-object-running-asyncio-coroutines-with-main-thread-input
#   https://stackoverflow.com/questions/32054066/python-how-to-run-multiple-coroutines-concurrently-using-asyncio
#   https://websockets.readthedocs.io/en/stable/intro.html
import os
import sys
import asyncio
import websockets
import janus
import queue
from msgtools.lib.messaging import Messaging

class Server:
    def __init__(self, timeout=10):
        self.loop = asyncio.get_event_loop()
        from threading import Thread
        self.t = Thread(target=self.start)
        self.t.start()
        self.timeout = timeout

    def send(self, msg):
        self.synchronous_tx_queue.sync_q.put(msg.rawBuffer().raw)
    
    def recv(self, msgIds=[], timeout=None):
        # if user didn't pass a list, put the single param into a list
        if not isinstance(msgIds, list):
            msgIds = [msgIds]
        # if they passed classes, get the ID of each
        for i in range(0,len(msgIds)):
            if hasattr(msgIds[i], 'ID'):
                msgIds[i] = msgIds[i].ID
        if timeout != None and self.timeout != timeout:
            self.timeout = timeout
        while True:
            try:
                data = self.synchronous_rx_queue.get(True, self.timeout)
                hdr = Messaging.hdr(data)
                id = hdr.GetMessageID()
                if len(msgIds) == 0 or id in msgIds:
                    msg = Messaging.MsgFactory(hdr)
                    return msg
            except queue.Empty:
                return None
    
    def stop(self):
        for task in asyncio.Task.all_tasks():
            task.cancel()
        self.tcp_server.close()
        self.loop.call_soon_threadsafe(self.loop.stop)
        self.t.join()

    # Implementation details from here down
    def start(self):
        # general asyncio stuff
        asyncio.set_event_loop(self.loop)
        self.loop = asyncio.get_event_loop()

        # synchronous input/output
        self.synchronous_tx_queue = janus.Queue(loop=self.loop)
        self.synchronous_rx_queue = queue.Queue()
        asyncio.ensure_future(self.handle_synchronous_input())

        # client lists
        self.tcp_clients = {} # task -> (reader, writer)
        self.ws_clients = {}

        # tcp server
        self.tcp_server = self.loop.run_until_complete(asyncio.start_server(self.client_connected_handler, '127.0.0.1', 5678))
        
        # websocket server
        start_ws_server = websockets.serve(self.handle_ws_client, '127.0.0.1', 5679)
        self.loop.run_until_complete(start_ws_server)
        
        # kick things off
        self.loop.run_forever()
        self.stopped()

    async def handle_synchronous_input(self):
        while True:
            data = await self.synchronous_tx_queue.async_q.get()
            await self.send_to_others(self.synchronous_tx_queue, data)

    def stopped(self):
        self.tcp_server.close()
        self.loop.stop()
        self.loop.close()
    
    async def send_to_others(self, me, data):
        if me != self.synchronous_tx_queue:
            try:
                #print("writing to synchronous_rx_queue")
                self.synchronous_rx_queue.put(data)
            except:
                pass

        # send to Websocket clients
        for ws in self.ws_clients.keys():
            if ws != me:
                # calling ws.send REQUIRES an 'await', otherwise data never gets to the ws client!
                await ws.send(data)

        # send to TCP clients
        for task in self.tcp_clients.keys():
                (reader, writer) = self.tcp_clients[task]
                if reader != me:
                    writer.write(data)

    async def handle_tcp_client(self, client_reader):
        #print("handle_tcp_client")
        while True:
            # we *should* read Messaging.hdr.SIZE bytes,
            # then parse header to see body length,
            # then read those bytes.
            data = await client_reader.read(1024)
            if not data:
                break
            await self.send_to_others(client_reader, data)

    def client_connected_handler(self, client_reader, client_writer):
        # Start a new asyncio.Task to handle this specific client connection
        task = asyncio.Task(self.handle_tcp_client(client_reader))
        print("Added new client to list of " + str(len(self.tcp_clients)) + " clients")
        self.tcp_clients[task] = (client_reader, client_writer)

        def client_done(task):
            print("client exited")
            # When the tasks that handles the specific client connection is done
            del self.tcp_clients[task]

        # Add the client_done callback to be run when the future becomes done
        task.add_done_callback(client_done)

    async def handle_ws_client(self, websocket, path):
        self.ws_clients[websocket] = websocket
        print("websocket connected")
        try:
            while True:
                data = await websocket.recv()
                if not data:
                    break
                await self.send_to_others(websocket, data)
        except websockets.exceptions.ConnectionClosed:
            print("websocket closed")
            del self.ws_clients[websocket]



from PyQt5 import QtCore, QtGui, QtWidgets, QtNetwork
from PyQt5.QtCore import QObject

from msgtools.lib.messaging import Messaging

class TcpClientConnection(QObject):
    disconnected = QtCore.pyqtSignal(object)
    messagereceived = QtCore.pyqtSignal(object)
    statusUpdate = QtCore.pyqtSignal(str)

    def __init__(self, tcpSocket):
        super(TcpClientConnection, self).__init__(None)

        self.removeClient = QtWidgets.QPushButton("Remove")
        self.removeClient.pressed.connect(lambda: self.tcpSocket.close())
        self.statusLabel = QtWidgets.QLabel()
        self.subscriptions = {}
        self.subMask = ~0
        self.subValue = 0
        self.isHardwareLink = False
        
        self.tcpSocket = tcpSocket
        self.tcpSocket.readyRead.connect(self.onReadyRead)
        self.tcpSocket.disconnected.connect(self.onDisconnected)

        self.rxBuffer = bytearray()

        self.name = "TCP Client"
        self.hostLabel = QtWidgets.QLabel(self.tcpSocket.peerAddress().toString().replace("::ffff:",""))
        self.statusLabel.setText(self.name)

    def widget(self, index):
        if index == 0:
            return self.removeClient
        if index == 1:
            return self.statusLabel
        if index == 2:
            return self.hostLabel
        return None
            
    def onReadyRead(self):
        inputStream = QtCore.QDataStream(self.tcpSocket)

        while(1):
            if len(self.rxBuffer) < Messaging.hdrSize:
                if self.tcpSocket.bytesAvailable() < Messaging.hdrSize:
                    return
                self.rxBuffer += inputStream.readRawData(Messaging.hdrSize - len(self.rxBuffer))

            if len(self.rxBuffer) >= Messaging.hdrSize:
                hdr = Messaging.hdr(self.rxBuffer)
                bodyLen = hdr.GetDataLength()
                if len(self.rxBuffer)+self.tcpSocket.bytesAvailable() < Messaging.hdrSize + bodyLen:
                    return

                self.rxBuffer += inputStream.readRawData(Messaging.hdrSize + bodyLen - len(self.rxBuffer))

                # create a new header object with the appended body
                hdr = Messaging.hdr(self.rxBuffer)

                # if we got this far, we have a whole message! So, emit the signal
                self.messagereceived.emit(hdr)

                # then clear the buffer, so we start over on the next message
                self.rxBuffer = bytearray()

    def onDisconnected(self):
        self.disconnected.emit(self)

    def sendMsg(self, msg):
        buf = msg.rawBuffer().raw
        while len(buf) > 0:
            bytesWritten = self.tcpSocket.write(buf)
            if bytesWritten == -1:
                raise Exception('Write error (-1)')
            buf = buf[bytesWritten:]
    
class TcpServer(QObject):
    statusUpdate = QtCore.pyqtSignal(str)
    newConnection = QtCore.pyqtSignal(object)

    def __init__(self, portNumber):
        super(TcpServer, self).__init__(None)

        self.portNumber = portNumber
        self.tcpServer = QtNetwork.QTcpServer()
        self.tcpServer.newConnection.connect(self.onNewTcpConnection)

    def start(self):
        if not self.tcpServer.listen(QtNetwork.QHostAddress.Any, self.portNumber):
            self.statusUpdate.emit("Con't open TCP socket on port "+str(self.portNumber)+"!")

    def onNewTcpConnection(self):
        connection = TcpClientConnection(self.tcpServer.nextPendingConnection())
        self.newConnection.emit(connection)
        connection.statusUpdate.connect(self.clientStatusUpdate)

    def serverInfo(self):
        # Show IP address and port number in status bar
        name = ""
        for address in QtNetwork.QNetworkInterface.allAddresses():
            if address.protocol() == QtNetwork.QAbstractSocket.IPv4Protocol and address != QtNetwork.QHostAddress.LocalHost:
                addrStr = address.toString()
                # ignore VM/VPN stuff (special ubuntu address, and anything that ends in 1)
                if ".".split(addrStr)[-1] != "1" and addrStr != "192.168.122.1":
                    name += address.toString() + "/"
        name = name[:-1]
        name += " : " + str(self.portNumber)
        return name

    def clientStatusUpdate(self, msg):
        self.statusUpdate.emit(msg)


import msgtools.parser.parser as MsgParser
from msgtools.parser.MsgUtils import *

def endian_string():
    if MsgParser.big_endian:
        return ''
    else:
        return ', Endian.little'

def fieldType(field):
    typeStr = field["Type"]
    if "int" in typeStr:
        return typeStr.title()
    if typeStr == "float32":
        return "Float32";
    if typeStr == "float64":
        return "Float64";
    return "?"
    
def returnType(field, bits):
    typeStr = field["Type"]
    if "int" in typeStr:
        if "Offset" in field or "Scale" in field or (bits and ("Offset" in bits or "Scale" in bits)):
            return "double";
        return "int"
    elif typeStr == "float32" or typeStr == "float64":
        return "double";
    return "?"
    
def fnHdr(field):
    ret = "/* %s %s, (%s to %s)*/" % (MsgParser.fieldDescription(field), MsgParser.fieldUnits(field), MsgParser.fieldMin(field), MsgParser.fieldMax(field))
    return ret

def getMath(x, field):
    ret = x
    if "Offset" in field or "Scale" in field:
        ret = "(%s).toDouble()" % (ret)
    if "Scale" in field:
        ret = "(%s * %s)" % (ret, fieldScale(field, ""))
    if "Offset" in field:
        ret = "(%s + %s)" % (ret, fieldOffset(field, ""))
    return ret

def setMath(x, field):
    ret = x
    if "Offset" in field:
        ret = "(%s - %s)" % (ret, fieldOffset(field, ""))
    if "Scale" in field:
        ret = "%s ~/ %s" % (ret, fieldScale(field, ""))
    if "Offset" in field or "Scale" in field:
        ret = "(%s).toInt()" % (ret)
    return ret


def getFn(field):
    loc = str(MsgParser.fieldLocation(field))
    param = ""
    if MsgParser.fieldCount(field) > 1:
        loc += "+idx*" + str(MsgParser.fieldArrayElementOffset(field))
        param += "int idx"
    access = "_data.get%s(%s%s)" % (fieldType(field), loc, endian_string())
    access = getMath(access, field)
    retType = returnType(field, None)
    #elif "Enum" in field:
    #    retType = field["Enum"]
    #    access = retType + "(" + access + ")"
    ret = '''\
%s
%s Get%s(%s)
{
    return %s;
}''' % (fnHdr(field), retType, field["Name"], param, access)
    if MsgParser.fieldUnits(field) == "ASCII" and (field["Type"] == "uint8" or field["Type"] == "int8"):
        ret += '''
%s
String Get%sString()
{
    final StringBuffer ret = new StringBuffer();
    for(int i=0; i<%s && i<_hdr.GetDataLength()-%s; i++)
    {
        final String nextChar = new String.fromCharCode(Get%s(i));
        if(nextChar == '\\0')
            break;
        ret.write(nextChar);
    }
    return ret.toString();
}''' % (fnHdr(field), field["Name"], str(MsgParser.fieldCount(field)), str(MsgParser.fieldLocation(field)), field["Name"])
    return ret

def setFn(field):
    paramType = returnType(field, None)
    valueString = setMath("value", field)
    #if "Enum" in field:
    #    valueString = paramType + "(" + valueString + ")"
    #    paramType = field["Enum"]
    param = paramType + " value"
    loc = str(MsgParser.fieldLocation(field))
    if MsgParser.fieldCount(field) > 1:
        loc += "+idx*" + str(MsgParser.fieldArrayElementOffset(field))
        param += ", int idx"
    ret = '''\
%s
void Set%s(%s)
{
    _data.set%s(%s, %s%s);
}''' % (fnHdr(field), field["Name"], param, fieldType(field), loc, valueString, endian_string())
    if MsgParser.fieldUnits(field) == "ASCII" and (field["Type"] == "uint8" or field["Type"] == "int8"):
        ret += '''
%s
void Set%sString(String value)
{
    for(int i=0; i<%s && i<value.length; i++)
    {
        Set%s(value.codeUnitAt(i), i);
    }
}''' % (fnHdr(field), field["Name"], str(MsgParser.fieldCount(field)), field["Name"])
    return ret

def getBitsFn(field, bits, bitOffset, numBits):
    access = "(Get%s() >> %s) & %s" % (field["Name"], str(bitOffset), MsgParser.Mask(numBits))
    access = getMath(access, bits)
    retType = returnType(field, bits)
    #if "Enum" in bits:
    #    retType = bits["Enum"]
    #    access = retType + "(" + access + ")"
    ret = '''\
%s
%s Get%s()
{
    return %s;
}''' % (fnHdr(bits), retType, MsgParser.BitfieldName(field, bits), access)
    return ret

def setBitsFn(field, bits, bitOffset, numBits):
    paramType = returnType(field, bits)
    valueString = setMath("value", bits)
    #if "Enum" in bits:
    #    valueString = paramType + "(" + valueString + ")"
    #    paramType = bits["Enum"]
    oldVal = '''%s() & ~(%s << %s)''' % ("Get"+field["Name"], MsgParser.Mask(numBits), str(bitOffset));
    newVal = '''(%s) | ((%s & %s) << %s)''' % (oldVal, valueString, MsgParser.Mask(numBits), str(bitOffset));
    ret = '''\
%s
void Set%s(%s value)
{
    %s(%s);
}''' % (fnHdr(bits), MsgParser.BitfieldName(field, bits), paramType, "Set"+field["Name"], newVal)
    return ret

def accessors(msg):
    gets = []
    sets = []
    
    if "Fields" in msg:
        for field in msg["Fields"]:
            gets.append(getFn(field))
            sets.append(setFn(field))
            bitOffset = 0
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    numBits = bits["NumBits"]
                    gets.append(getBitsFn(field, bits, bitOffset, numBits))
                    sets.append(setBitsFn(field, bits, bitOffset, numBits))
                    bitOffset += numBits

    return gets+sets

def fieldDefault(field):
    ret = field["Default"]
    return ret

def initField(field):
    if "Default" in field:
        if MsgParser.fieldCount(field) > 1:
            ret = "for(int i=0; i<" + str(MsgParser.fieldCount(field)) + "; i++)\n"
            ret += "    Set" + field["Name"] + "(" + str(fieldDefault(field)) + ", i);" 
            return ret;
        else:
            return  "Set" + field["Name"] + "(" + str(fieldDefault(field)) + ");"
    return ""

def initBitfield(field, bits):
    if "Default" in bits:
        return  "Set" + MsgParser.BitfieldName(field, bits) + "(" + str(bits["Default"]) + ");"
    return ""

def initCode(msg):
    ret = []
    if "Fields" in msg:
        for field in msg["Fields"]:
            fieldInit = initField(field)
            if fieldInit:
                ret.append(fieldInit)
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    bits = initBitfield(field, bits)
                    if bits:
                        ret.append(bits)

    return ret

def enums(e):
    ret = ""
    for enum in e:
        ret += "class <MSGNAME>_"+enum["Name"]+" {"
        for option in enum["Options"]:
            optionName = OptionName(option)
            ret += "static const int "+optionName+" = "+str(option["Value"]) + ';'
        ret += "}\n"
    return ret

def fieldReflectionType(field):
    ret = fieldType(field)
    if field["Type"] == "float32" or field["Type"] == "float64":
        return "FloatFieldInfo"

    if ret.lower().startswith("int"):
        ret = "IntFieldInfo"
    if ret.lower().startswith("uint"):
        ret = "UIntFieldInfo"

    if "NumBits" in field:
        ret = "BitfieldInfo"
        if "Offset" in field or "Scale" in field:
            ret = "ScaledBitfieldInfo"
    else:
        if "Offset" in field or "Scale" in field:
            ret = "ScaledFieldInfo"
    if "Enum" in field:
        ret = "EnumFieldInfo"
    return ret

def fieldReflectionBitsType(field, bits):
    ret = fieldType(field)
    if field["Type"] == "float32" or field["Type"] == "float64":
        return "FloatFieldInfo"

    if ret.lower().startswith("int"):
        ret = "IntFieldInfo"
    if ret.lower().startswith("uint"):
        ret = "UIntFieldInfo"

    if "NumBits" in bits:
        ret = "BitfieldInfo"
        if "Offset" in bits or "Scale" in bits:
            ret = "ScaledBitfieldInfo"
    else:
        if "Offset" in field or "Scale" in field:
            ret = "ScaledFieldInfo"
    if "Enum" in field:
        ret = "EnumFieldInfo"
    return ret

def fieldReflection(field):
    loc = str(MsgParser.fieldLocation(field))
    type = fieldReflectionType(field)
    params = type;
    params += "("
    params += '"'+field["Name"] + '"'
    params += ', "' + MsgParser.fieldDescription(field) + '"'
    params += ', "' + MsgParser.fieldUnits(field) + '"'
    params += ", " + loc
    params += ", " + str(MsgParser.fieldSize(field))
    params += ", " + str(MsgParser.fieldCount(field))
    if "Offset" in field or "Scale" in field:
        if "Scale" in field:
            params += ", " + str(field["Scale"])
        else:
            params += ", 1.0"
        if "Offset" in field:
            params += ", " + str(field["Offset"])
        else:
            params += ", 0.0"
    params += ")"
    return params

def fieldBitsReflection(field, bits, bitOffset, numBits):
    loc = str(MsgParser.fieldLocation(field))
    type = fieldReflectionBitsType(field, bits)
    params = type;
    params += "("
    params += '"'+bits["Name"] + '"'
    params += ', "' + MsgParser.fieldDescription(bits) + '"'
    params += ', "' + MsgParser.fieldUnits(bits) + '"'
    params += ", " + loc
    params += ", " + str(MsgParser.fieldSize(field))
    params += ", " + str(MsgParser.fieldCount(bits))
    if "Offset" in bits or "Scale" in bits:
        if "Scale" in bits:
            params += ", " + str(bits["Scale"])
        else:
            params += ", 1.0"
        if "Offset" in bits:
            params += ", " + str(bits["Offset"])
        else:
            params += ", 0.0"
    params += ", " + str(bitOffset)
    params += ", " + str(numBits)
    params += ")"
    return params

def reflection(msg):
    ret = []
    
    if "Fields" in msg:
        for field in msg["Fields"]:
            ret.append(fieldReflection(field))
            bitOffset = 0
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    numBits = bits["NumBits"]
                    ret.append(fieldBitsReflection(field, bits, bitOffset, numBits))
                    bitOffset += numBits

    return "\n".join(ret)

def fieldMin(field):
    ret = str(MsgParser.fieldMin(field))
    if ret == 'DBL_MIN':
        ret = '-double.MAX_FINITE'
    elif ret == 'FLT_MIN':
        ret = '-3.402823466e38'
    return ret

def fieldMax(field):
    ret = str(MsgParser.fieldMax(field))
    if ret == 'DBL_MAX':
        ret = 'double.MAX_FINITE'
    elif ret == 'FLT_MAX':
        ret = '3.402823466e38'
    return ret

def genericInfo(field, fieldName, loc, fieldType):
    params  = 'static const int '+fieldName+'_Loc   = ' + loc + ';\n'
    params += 'static const '+fieldType+" "+fieldName+'_Max   = ' + fieldMax(field) + ';\n'
    params += 'static const '+fieldType+" "+fieldName+'_Min   = ' + fieldMin(field) + ';\n'
    params += 'static const String '+fieldName+"_Units = '" + str(MsgParser.fieldUnits(field)) + "'" + ';\n'
    params += 'static const int '+fieldName+'_Count = ' + str(MsgParser.fieldCount(field)) + ';\n'
    if "Default" in field:
        params += 'static const '+fieldType+" "+fieldName+'_DefaultValue = ' + str(fieldDefault(field)) + ";\n" 
    if "Scale" in field:
        params += 'static const double '+fieldName+'_Scale = ' + str(field["Scale"]) + ';\n'
    if "Offset" in field:
        params += 'static const double '+fieldName+'_Offset = ' + str(field["Offset"]) + ';\n'
    return params
    
def fieldInfo(field):
    params = genericInfo(field, field["Name"], str(MsgParser.fieldLocation(field)), returnType(field,None))
    return params

def fieldBitsInfo(field, bits, bitOffset, numBits):
    params  = genericInfo(bits, MsgParser.BitfieldName(field, bits), str(MsgParser.fieldLocation(field)), returnType(field,bits))
    params += 'static const int '+MsgParser.BitfieldName(field, bits)+'_BitOffset = ' + str(bitOffset) + ';\n'
    params += 'static const int '+MsgParser.BitfieldName(field, bits)+'_NumBits   = ' + str(numBits) + ';\n'
    return params

def fieldInfos(msg):
    ret = []
    
    if "Fields" in msg:
        for field in msg["Fields"]:
            ret.append(fieldInfo(field))
            bitOffset = 0
            if "Bitfields" in field:
                for bits in field["Bitfields"]:
                    numBits = bits["NumBits"]
                    ret.append(fieldBitsInfo(field, bits, bitOffset, numBits))
                    bitOffset += numBits

    return "\n".join(ret)

def declarations(msg):
    return []

def getMsgID(msg):
    ret = baseGetMsgID("", "", 0, 0, msg)
    return ret.replace("uint32_t", "int")
    
def setMsgID(msg):
    return baseSetMsgID("", "", 0, 0, msg)



from .messaging import Messaging

# for conversion to JSON
from collections import OrderedDict
import json

def toJson(msg, includeHeader=False):
    pythonObj = OrderedDict()
    if includeHeader:
        pythonObj['hdr'] = OrderedDict()
        for fieldInfo in msg.hdr.fields:
            if(fieldInfo.count == 1):
                if len(fieldInfo.bitfieldInfo) == 0:
                    pythonObj['hdr'][fieldInfo.name] = str(Messaging.get(msg.hdr, fieldInfo))
                else:
                    for bitInfo in fieldInfo.bitfieldInfo:
                        pythonObj['hdr'][bitInfo.name] = str(Messaging.get(msg.hdr, bitInfo))
            else:
                arrayList = []
                terminate = 0
                for i in range(0,fieldInfo.count):
                    arrayList.append(str(Messaging.get(msg.hdr, fieldInfo, i)))
                pythonObj['hdr'][fieldInfo.name] = arrayList

    msgClass = Messaging.MsgClass(msg.hdr)
    for fieldInfo in msgClass.fields:
        if(fieldInfo.count == 1):
            #TODO Broken for arrays-of-structs acting like parallel arrays!
            if not fieldInfo.exists(msg):
                break
            if len(fieldInfo.bitfieldInfo) == 0:
                pythonObj[fieldInfo.name] = str(Messaging.get(msg, fieldInfo))
            else:
                for bitInfo in fieldInfo.bitfieldInfo:
                    pythonObj[bitInfo.name] = str(Messaging.get(msg, bitInfo))
        else:
            arrayList = []
            terminate = 0
            for i in range(0,fieldInfo.count):
                #TODO Broken for arrays-of-structs acting like parallel arrays!
                if not fieldInfo.exists(msg, i):
                    terminate = 1
                    break
                arrayList.append(str(Messaging.get(msg, fieldInfo, i)))
            pythonObj[fieldInfo.name] = arrayList
            if terminate:
                break

    return json.dumps({msg.MsgName() : pythonObj})

def jsonToMsg(jsonString):
    terminationLen = None
    if "hdr" in jsonString:
        fieldJson = jsonString["hdr"]
        for fieldName in fieldJson:
            if fieldName == "DataLength":
                if fieldJson[fieldName] == ";":
                    terminationLen = 0
                else:
                    terminationLen = int(fieldJson[fieldName])
    for msgName in jsonString:
        if msgName == "hdr":
            # hdr handled above, *before* message body
            pass
        else:
            fieldJson = jsonString[msgName]
            msgClass = Messaging.MsgClassFromName[msgName]
            msg = msgClass()
            for fieldName in fieldJson:
                fieldInfo = Messaging.findFieldInfo(msgClass.fields, fieldName)
                fieldValue = fieldJson[fieldName]
                if isinstance(fieldValue, list):
                    #print(fieldName + " list type is " + str(type(fieldValue)))
                    for i in range(0,len(fieldValue)):
                        Messaging.set(msg, fieldInfo, fieldValue[i], i)
                        if terminationLen != None:
                            #TODO Broken for arrays-of-structs acting like parallel arrays!
                            terminationLen = max(terminationLen, fieldInfo.end_location(i))
                elif isinstance(fieldValue, dict):
                    #print(fieldName + " dict type is " + str(type(fieldValue)))
                    if fieldInfo.bitfieldInfo:
                        pass
                    else:
                        pass
                else:
                    #print(str(type(fieldValue)) + " " + fieldName + ", calling set with " + str(fieldValue))
                    Messaging.set(msg, fieldInfo, fieldValue)
                    if terminationLen != None:
                        if fieldInfo.type == "string":
                            #TODO Broken for arrays-of-structs acting like parallel arrays!
                            terminationLen = max(terminationLen, fieldInfo.end_location(len(fieldValue)-1))
                        else:
                            #TODO Broken for arrays-of-structs acting like parallel arrays!
                            terminationLen = max(terminationLen, fieldInfo.end_location())
    if terminationLen != None:
        msg.hdr.SetDataLength(terminationLen)
    return msg



#!/usr/bin/env python3
import re
import sys
import os
import string
import collections
import hashlib

def Usage():
    sys.stderr.write('Usage: ' + sys.argv[0] + ' inputdir dictionaryFile dictionaryHeaderFile\n')
    sys.exit(1)

def ProcessDir(inputDir):
    for filename in sorted(os.listdir(inputDir)):
        inputFilename = inputDir + '/' + filename
        if os.path.isdir(inputFilename):
            ProcessDir(inputFilename)
        else:
            if filename.endswith(".h") or filename.endswith(".c") or filename.endswith(".cpp"):
                if filename != "debug_printf.h" and filename != "debug_printf.cpp":
                    #print("reading " + inputFilename)
                    ProcessFile(inputFilename.replace("//", "/"))

def printDictionary(dictFilename, headerFilename):
    md5 = hashlib.md5()
    with open(dictFilename,'w') as dictFile:
        header = \
'''# AUTO-GENERATED FILE, DO NOT HAND EDIT!
# Created by %s
''' % thisInvocation
        dictFile.write(header)
        for formatInfo in dictionary:
            s = str(formatInfo.id) + ": " + formatInfo.formatStr+", "+formatInfo.filename+", " + str(formatInfo.linenumber) + "\n"
            md5.update(s.encode('utf-8'))
            dictFile.write(s)
        md5 = md5.hexdigest()
        dictFile.write("Dictionary md5 is " + md5)

    lineEndings = os.linesep
    # for windows, override lineseperator, to force windows native line endings,
    # even if we're running cygwin with \n line endings.
    if sys.platform.startswith('win32') or sys.platform.startswith('cygwin'):
        lineEndings="\r\n"

    with open(headerFilename,'w', newline=lineEndings) as hdrFile:
        header = \
'''/* AUTO-GENERATED FILE, DO NOT HAND EDIT!
  Created by %s
  */
''' % thisInvocation
        hdrFile.write(header)
        for formatInfo in dictionary:
            formatStrIdName = formatInfo.filename.replace("/","_").replace(".cpp","").replace(".c","").replace(".","_")+"_line_"+str(formatInfo.linenumber)
            hdrFile.write("#define " + formatStrIdName + " " + str(formatInfo.id)+"\n")
        md5array = "0x" + ", 0x".join([md5[i:i+2] for i in range(0, len(md5), 2)])
        hdrFile.write("#define FORMAT_STR_DICTIONARY_ID " + md5array+"\n")

PrintfInfo = collections.namedtuple('PrintfInfo', ['id', 'formatStr', 'filename', 'linenumber'])

def ProcessFile(inputFilename):
    try:
        escapedFilePath = inputFilename.replace("/","_").replace(".cpp","").replace(".c","").replace(".","_")
        with open(inputFilename, 'r') as inputFile:
            inputData = inputFile.read().splitlines() 
        global maxSubscriptions
        global formatStrId
        lineNumber = 0;
        for line in inputData:
            lineNumber+=1
            formatStr = ""
            try:
                printStatement = ""
                if "#define ESCAPED_FILE_PATH" in line:
                    matchObj = re.search( r'#define\s*ESCAPED_FILE_PATH\s*(.*)', line)
                    escapedFilePathFromFile = matchObj.group(1).strip()
                    if escapedFilePath != escapedFilePathFromFile:
                        print(inputFilename + ":"+str(lineNumber) + ", " + escapedFilePathFromFile + " != " + escapedFilePath)
                        sedCmd = "sed -i '"+str(lineNumber)+"s/"+escapedFilePathFromFile+"/"+escapedFilePath+"/' "+inputFilename
                        #print("need to " + sedCmd)
                        os.system(sedCmd)
                if "debugPrintf(" in line:
                    printStatement = "debugPrintf"
                    matchObj = re.search( r'debugPrintf\(.*("[^"]*")', line)
                    formatStr = matchObj.group(1).strip()
                if "debugWarn(" in line:
                    printStatement = "debugWarn"
                    matchObj = re.search( r'debugWarn\(.*("[^"]*")', line)
                    formatStr = matchObj.group(1).strip()
                if "debugError(" in line:
                    printStatement = "debugError"
                    matchObj = re.search( r'debugError\(.*("[^"]*")', line)
                    formatStr = matchObj.group(1).strip()
                if(formatStr != ""):
                    printInfo = PrintfInfo(formatStrId, formatStr, inputFilename, lineNumber)
                    dictionary.append(printInfo)
                    formatStrId+=1
            except AttributeError:
                print("Regex search error on " + inputFilename + ", line " + str(lineNumber))
                print(line)
                sys.exit(1)
    except UnicodeDecodeError:
        print("UnicodeDecodeError on " + inputFilename) 

def main():
    if len(sys.argv) < 4:
        Usage();
    global thisInvocation
    thisInvocation = " ".join(sys.argv)
    inputDir = sys.argv[1]
    dictionaryFile = sys.argv[2]
    dictionaryHeaderFile = sys.argv[3]

    global dictionary
    dictionary = []
    
    global formatStrId
    formatStrId = 0
    
    inputDirList = inputDir.split(',')
    for d in inputDirList:
        ProcessDir(d)

    try:
        os.makedirs(os.path.dirname(dictionaryFile))
    except:
        pass
    printDictionary(dictionaryFile, dictionaryHeaderFile)

# main starts here
if __name__ == '__main__':
    main()
